#include "..\Precomp.h"
#include "Ammo.h"
#include "..\Network\Network.h"

/* Global */
AmmoFactory g_AmmoFact;

/* AmmoBase */
void AmmoBase::NetHit(ShipObj *s, Object *o)
{ if(!g_World.bNet || !s) return;
  if(o->ID)
  { MsgObjectHit msg;
    msg.ID = Ship->ID, msg.HitID = s->ID, msg.ObjID = o->ID;
    msg.X  = o->Pos.X, msg.Y  = o->Pos.Y;
    msg.XV = o->Vel.X, msg.YV = o->Vel.Y;
    g_World.Send(&msg, NET_SENDGUAR);
  }
  else
  { MsgFWeaponHit msg;
    msg.ID = Ship->ID, msg.HitID = s->ID, msg.AType = AType;
    msg.X  = o->Pos.X, msg.Y  = o->Pos.Y;
    msg.XV = o->Vel.X, msg.YV = o->Vel.Y;
    msg.Weight = (WORD)o->Weight, msg.Extra = (WORD)Extra;
    msg.Color  = o->Color;
    g_World.Send(&msg, NET_SENDGUAR);
  }
}

void AmmoBase::NetKill(Object *o)
{ if(g_World.bNet && o->ID && Ship==&g_World.Me->Ship)
  { MsgObjectHit msg;
    msg.ID = Ship->ID, msg.ObjID = o->ID, msg.HitID = 0;
    msg.X  = o->Pos.X, msg.Y  = o->Pos.Y;
    msg.XV = o->Vel.X, msg.YV = o->Vel.Y;
    g_World.Send(&msg, NET_SENDGUAR);
  }
}

/* AmmoBullet */
AmmoBullet::AmmoBullet()
{ Weight = 3;
  Extra  = 6;
  Color  = RC_WHITE;
}

void AmmoBullet::Draw(int x, int y)
{ Plot2D(g_Back, (int)Pos.X-x, (int)Pos.Y-y, Color);
}


/* AmmoCannonBall */
AmmoCannonBall::AmmoCannonBall()
{ Weight=35;
  Extra =35;
  Radius=9;
}

void AmmoCannonBall::Draw(int x, int y)
{ RECT src = SPR_CANNONBALL;
  g_Sprites->BlitFast(g_Back, (int)Pos.X-x-(src.right-src.left)/2,
                      (int)Pos.Y-y-(src.bottom-src.top)/2, &src, CDD_CLIP);
}

void AmmoCannonBall::Damage(ShipObj *s)
{ s->Impact(this, Extra);
  g_World.MkExp(ET_TINY, (int)Pos.X, (int)Pos.Y, Ship);
}

void AmmoCannonBall::HitMap()
{ g_World.MkExp(ET_TINY, (int)Pos.X, (int)Pos.Y, Ship);
  Flag.bDead = true;
}


/* AmmoGrenade */
AmmoGrenade::AmmoGrenade()
{ Weight=10;
  Extra =10;
  Radius=8;
}

void AmmoGrenade::Draw(int x, int y)
{ RECT src = SPR_GRENADE;
  g_Sprites->BlitFast(g_Back, (int)Pos.X-x-(src.right-src.left)/2,
                      (int)Pos.Y-y-(src.bottom-src.top)/2, &src, CDD_CLIP);
}

void AmmoGrenade::Tick(unsigned)
{ if(Age > TICKS_PER_SEC/2)
  { AmmoBullet *b;
    Flag.bDead = true;
    for(int a=0; a<256; a+=rand()%8)
    { for(int n=0; n<2; n++)
      { b = (AmmoBullet*)g_AmmoFact.Create(AT_BULLET);
        b->Color  = Ship->Color-rand()%3;
        b->Extra  = 3;
        b->Vel    = Vel+g_Vects[a]*(rand()%256/512.0);
        b->Pos    = Pos+g_Vects[a]*(rand()%256/64.0);
        b->LoadShip(Ship);
        g_World.AddObject(b);
      }
    }
  }
}

void AmmoGrenade::HitMap()
{ g_World.MkExp(ET_SMALL, (int)Pos.X, (int)Pos.Y, Ship);
  Flag.bDead = true;
}


/* AmmoDropBomb */
AmmoDropBomb::AmmoDropBomb()
{ Weight=8;
  Extra =12;
  Radius=9;
}

void AmmoDropBomb::Draw(int x, int y)
{ RECT src = SPR_DROPBOMB;
  g_Sprites->BlitFast(g_Back, (int)Pos.X-x-(src.right-src.left)/2,
                      (int)Pos.Y-y-(src.bottom-src.top)/2, &src, CDD_CLIP);
}

void AmmoDropBomb::HitMap()
{ g_World.MkExp(ET_TINY, (int)Pos.X, (int)Pos.Y, Ship);
  Flag.bDead = true;
}


/* AmmoDynamite */
AmmoDynamite::AmmoDynamite()
{ Weight=1;
  Radius=40;
  Extra =50;
  Flag.bNoGrav=true;
}

void AmmoDynamite::Draw(int x, int y)
{ RECT src = SPR_DYNAMITE;
  g_Sprites->BlitFast(g_Back, (int)Pos.X-x-(src.right-src.left)/2,
                      (int)Pos.Y-y-(src.bottom-src.top)/2, &src, CDD_CLIP);
}

void AmmoDynamite::Damage(ShipObj *s)
{ s->Impact(this, Extra);
  g_World.MkExp(ET_LARGE, (int)Pos.X, (int)Pos.Y, Ship);
}

void AmmoDynamite::Tick(unsigned)
{ if(Age >= TICKS_PER_SEC*3/2)
  { Flag.bDead=true;
    g_World.MkExp(ET_LARGE, (int)Pos.X, (int)Pos.Y, Ship);
  }
  else if((Age & 3) == 0)
  { static byte Colors[3] = { RC_RED, RC_YELLOW, RC_ORANGE };
    RECT src = SPR_DYNAMITE;
    SputterObj *s = SputterObj::Alloc.New();
    s->Pos  = Vector<FLOAT>((src.right-src.left)/2-1, -(src.bottom-src.top)/2+1)+Pos;
    s->Vel  = g_Vects[(byte)(rand()%64)]*0.5;
    s->Life = rand()%8+6;
    s->Color = Colors[rand()%3];
    g_World.AddObject(s);
  }
}

/* AmmoPoint */
void AmmoPoint::LoadFromMsg(MsgWeaponFire *msg)
{ int i = g_World.FindPlayer(msg->ID);
  assert(i != -1);

  Pos.X  = msg->X,  Pos.Y = msg->Y;
  Vel.X  = msg->XV, Vel.Y = msg->YV;
  ID     = msg->ObjID,  Color = msg->Color;
  Weight = msg->Weight, Extra = msg->Extra;
  LoadShip(&g_World.Players[i]->Ship);
  if(msg->AmmoData) LoadMsgData(msg);
}

bool AmmoPoint::Intersects(Object *o)
{ bool hit = PointObj::Intersects(o);
  if(o==Ship)
  { if(bGrace && (bGrace=hit)) hit=false;
  }
  return hit;
}

void AmmoPoint::HitShip(ShipObj *s)
{ Flag.bDead=true;
  if(Ship==&g_World.Me->Ship)
  { Damage(s);
    NetHit(s, this);
  }
}


/* AmmoSphere */
void AmmoSphere::LoadFromMsg(MsgWeaponFire *msg)
{ int i = g_World.FindPlayer(msg->ID);
  assert(i != -1);

  Pos.X  = msg->X,      Pos.Y  = msg->Y;
  Vel.X  = msg->XV,     Vel.Y  = msg->YV;
  Weight = msg->Weight, Extra  = msg->Extra;
  ID     = msg->ObjID,  Color  = msg->Color;
  Radius = msg->Radius;

  LoadShip(&g_World.Players[i]->Ship);
  if(msg->AmmoData) LoadMsgData(msg);
}

bool AmmoSphere::Intersects(Object *o)
{ bool hit = SphereObj::Intersects(o);
  if(o==Ship && bGrace && (bGrace=hit)) hit=false;
  return hit;
}

void AmmoSphere::HitShip(ShipObj *s)
{ Flag.bDead=true;
  if(Ship==&g_World.Me->Ship)
  { Damage(s);
    NetHit(s, this);
  }
}


/* AmmoFactory */
AmmoFactory::AmmoFactory()
{ assert(AT_NUM==5);
}

void AmmoFactory::Init()
{ Ammos[AT_BULLET]     = AmmoBullet::Alloc.New();
  Ammos[AT_CANNONBALL] = NEW AmmoCannonBall;
  Ammos[AT_GRENADE]    = NEW AmmoGrenade;
  Ammos[AT_DROPBOMB]   = NEW AmmoDropBomb;
  Ammos[AT_DYNAMITE]   = NEW AmmoDynamite;
}

void AmmoFactory::Deinit()
{ for(int i=0; i<sizeof(Ammos)/sizeof(AmmoBase*); i++)
    Ammos[i]->ToObject()->Destroy();
}

AmmoBase * AmmoFactory::Create(AmmoType type)
{ AmmoBase *ret=Ammos[type]->Clone();
  ret->AType = type;
  return ret;
}

AmmoBase * AmmoFactory::Static(AmmoType type)
{ AmmoBase *ret=Ammos[type];
  ret->AType = type;
  return ret;
}


/* statics */
BlkAlloc<AmmoBullet> AmmoBullet::Alloc(256);


