#ifndef AMMO_H_INC
#define AMMO_H_INC

enum AmmoType
{ AT_BULLET, AT_CANNONBALL, AT_GRENADE, AT_DROPBOMB, AT_DYNAMITE, AT_NUM
};

#ifndef TCSERVER

#include "GameObj.h"
#include "Ship.h"
#include "Utility\BlkAlloc.h"

struct MsgWeaponFire;

class AmmoBase
{ public:
    AmmoBase() { Ship=NULL; bGrace=true; }

    virtual void LoadFromMsg(MsgWeaponFire *) = 0;
    virtual void LoadMsgData(MsgWeaponFire *) { };
    virtual void LoadShip(ShipObj *s) { Ship=s; ToObject()->Owner=s->ID; }
    virtual AmmoBase * Clone() = 0;
    virtual Object * ToObject()= 0;
    void    NetHit(ShipObj *s, Object *);
    void    NetKill(Object *);

    ShipObj *Ship;
    int      Extra;
    AmmoType AType;
    bool     bGrace;
};

class AmmoPoint : public AmmoBase, public PointObj
{ public:
    virtual void LoadFromMsg(MsgWeaponFire *);
    virtual Object * ToObject() { return this; }
    virtual void Damage(ShipObj *s) { s->Impact(this, Extra); }
    virtual void HitMap() { Flag.bDead=true; NetKill(this); }
    virtual void HitShip(ShipObj *);
    virtual bool Intersects(Object *);
    virtual void Destroy()     { delete this; }
};

class AmmoSphere : public AmmoBase, public SphereObj
{ public:
    virtual void LoadFromMsg(MsgWeaponFire *);
    virtual Object * ToObject() { return this; }
    virtual void Damage(ShipObj *s) { s->Impact(this, Extra); }
    virtual void HitMap() { Flag.bDead=true; NetKill(this); }
    virtual void HitShip(ShipObj *);
    virtual bool Intersects(Object *);
    virtual void Destroy()     { delete this; }
};

class AmmoBullet : public AmmoPoint
{ public:
    AmmoBullet();

    virtual void Draw(int, int);
    virtual void Destroy()     { Alloc.Delete(this); }
    virtual AmmoBase * Clone() { return Alloc.New(); }

    static GameLib::BlkAlloc<AmmoBullet> Alloc;
};

class AmmoCannonBall : public AmmoSphere
{ public:
    AmmoCannonBall();
    virtual void Draw(int, int);
    virtual void Damage(ShipObj *s);
    virtual void HitMap();
    virtual AmmoBase * Clone() { return NEW AmmoCannonBall; }
};

class AmmoGrenade : public AmmoSphere
{ public:
    AmmoGrenade();
    virtual void Draw(int, int);
    virtual void Tick(unsigned);
    virtual void HitMap();
    virtual AmmoBase * Clone() { return NEW AmmoGrenade; }
};

class AmmoDropBomb : public AmmoSphere
{ public:
    AmmoDropBomb();
    virtual void Draw(int, int);
    virtual void HitMap();
    virtual AmmoBase * Clone() { return NEW AmmoDropBomb; }
};

class AmmoDynamite : public AmmoSphere
{ public:
    AmmoDynamite();
    virtual void Draw(int, int);
    virtual void Tick(unsigned);
    virtual void Damage(ShipObj *s);
    virtual void HitMap()      { g_World.MkExp(ET_LARGE, (int)Pos.X, (int)Pos.Y, Ship); Flag.bDead=true; }
    virtual AmmoBase * Clone() { return NEW AmmoDynamite; }
};

class AmmoFactory
{ public:
    AmmoFactory();

    void Init();
    void Deinit();
    AmmoBase * Create(AmmoType type);
    AmmoBase * Static(AmmoType type);

  protected:
    AmmoBase *Ammos[AT_NUM];
};

#endif // ndef TCSERVER

#endif

