#include "..\Precomp.h"
#include "Font.h"
#include "..\Utils\VFS.h"
#include "Genfile\memfile.h"
#include "Utility\DataLoad.h"

/*** GLOBALS ***/
CFont g_Fonter;
Font  g_Font;

/*** Font IMPLEMENTATION ***/
void Font::StringDims(const char *str, int len, int *wid, int *hei)
{ assert(Surf.GetDDS());
  int i=0, width=0, height=0, t, c;
  while(i<len)
  { c = str[i++] - 32;
    if(c<0 || c>94) continue;
    width += Places[c].right - Places[c].left;
    t = Places[c].bottom - Places[c].top;
    if(t>height) height = t;
  }
  if(wid) *wid = width;
  if(hei) *hei = height;
}

void Font::StringDimsF(int *wid, int *hei, const char *str, ...)
{ char buf[1024];
  va_list list;
  va_start(list, str);
  vsprintf(buf, str, list);
  va_end(list);
  StringDims(buf, wid, hei);
}

void Font::Print(CDDSurface *dest, const char *str, int x, int y, CDDColor c)
{ RECT clip = { x, y, dest->Width, dest->Height };
  Print(dest, str, &clip, c);
}

void Font::Print(CDDSurface *dest, const char *str, RECT *clip, CDDColor c)
{ assert(Surf.GetDDS());
  RECT  dr, *fr;
  int   ch, cwid;

  dr.left = clip->left;
  dr.top  = clip->top;
  if(dr.top+FontHei > clip->bottom) return;

  while(*str)
  { ch = *str++ - 32;
    if(ch<0 || ch>94) continue;
    fr = Places+ch;
    if(dr.left+(cwid=fr->right-fr->left) > clip->right)
    { RECT rect = { fr->left, fr->top, fr->left-dr.left+clip->right,
                    fr->top-dr.top+clip->bottom };
      dr.right = clip->right, dr.bottom = clip->bottom;
      dest->AlphaFill(&dr, c, 0, &Surf, &rect);
    }
    else
    { dr.right = dr.left+cwid, dr.bottom = dr.top+fr->bottom-fr->top;
      dest->AlphaFill(&dr, c, 0, &Surf, fr);
    }
    dr.left += cwid;
  }
}

void Font::PrintF(CDDSurface *dest, int x, int y, CDDColor c, const char *str, ...)
{ RECT clip = { x, y, dest->Width, dest->Height };
  char buf[1024];
  va_list list;
  va_start(list, str);
  vsprintf(buf, str, list);
  va_end(list);
  Print(dest, buf, &clip, c);
}

void Font::PrintF(CDDSurface *dest, RECT *clip, CDDColor c, const char *str, ...)
{ char buf[1024];
  va_list list;
  va_start(list, str);
  vsprintf(buf, str, list);
  va_end(list);
  Print(dest, buf, clip, c);
}

/*** CFont IMPLEMENTATION ***/
CFont::CFont()
{ Cddraw = NULL;
  VFS    = NULL;
}

void CFont::Init(CDDraw *cdd, CVFS *vfs)
{ Cddraw = cdd, VFS = vfs;
}

int CFont::LoadFont(const TCHAR *name, Font *font)
{ font->Surf.Deinit();
  memfile file;
  int      max=0, wid=0;

  if(VFS->LoadDataB(name, &file)) return 1;
  file.read(font->Places, 95 * sizeof(RECT));
  for(int i=0;i<95;i++)
  { if(font->Places[i].bottom-font->Places[i].top>max)
      max  = font->Places[i].bottom-font->Places[i].top;
      wid += font->Places[i].right-font->Places[i].left;
  }
  font->FontHei=max, font->AvgWid=(int)((float)wid/95+0.5);

  GameLib::Loader.SetCDDraw(&g_DD);
  if(!GameLib::Loader.LoadPcx(&file, &font->Surf, CDD_SYSTEM, CDD_8BIT))
    return 2;
  font->Surf.Stats.Trans = true;

  return 0;
}
