#include "..\Precomp.h"
#include "GameObj.h"

#include "..\Network\Network.h"

/* Globals */
ObjFactory g_ObjFact;

/* Object */
Object::Object()
{ Weight    = Flags = 0;
  Angle     = 0;
  BoundType = OBJ_UNKNOWN;
  Index     = 0x80000000;
  ID        = Owner = Age = 0;
}

void Object::CalcVel()
{ if(!Flag.bNoGrav) AddForce(0, g_World.Gravity);
}


/* PointObj */
bool PointObj::Intersects(Object *o)
{ switch(o->BoundType)
  { case OBJ_POINT:  return Round(Pos.X)==Round(o->Pos.X) &&
                            Round(Pos.Y)==Round(o->Pos.Y);
    case OBJ_SPHERE:
    { FLOAT xd=o->Pos.X-Pos.X, yd=o->Pos.Y-Pos.Y, dist=xd*xd+yd*yd;
      return dist <= ((SphereObj*)o)->Radius;
    }
  }
  assert(0);
  return false;
}


/* SphereObj */
bool SphereObj::Intersects(Object *o)
{ FLOAT xd=o->Pos.X-Pos.X, yd=o->Pos.Y-Pos.Y, dist=xd*xd+yd*yd;
  switch(o->BoundType)
  { case OBJ_POINT:  return dist <= Radius;
    case OBJ_SPHERE: return dist <= max(Radius, ((SphereObj*)o)->Radius);
  }
  assert(0);
  return false;
}


/* SputterObj */
void SputterObj::Draw(int x, int y)
{ Plot2D(g_Back, Round(Pos.X)-x, Round(Pos.Y)-y, Color);
}


/* ExplodeObj */
ExplodeObj::ExplodeObj()
{ HitCounter   = Activate = Stage = 0;
  Flag.bNoGrav = true;
}

void ExplodeObj::HitShip(ShipObj *s)
{ static damage[3] = { 40, 50, 60 };
  static weight[3] = { 15, 25, 40 };
  if(!HitCounter && !Activate)
  { Weight = weight[Type];
    s->Impact(this, damage[Type]);
    HitCounter=TICKS_PER_SEC/2;
  }
}

void ExplodeObj::Draw(int x, int y)
{ if(Activate) return;
  RECT src = Types[Type].Rects[Stage];
  g_Sprites->BlitFast(g_Back, (int)Pos.X-x-(src.right-src.left)/2,
                      (int)Pos.Y-y-(src.bottom-src.top)/2, &src, CDD_CLIP);
}

void ExplodeObj::Tick(unsigned)
{ if(Age < Activate) return;
  else if(Activate) Age=Activate=0;

  if(HitCounter) HitCounter--;
  int next = Types[Type].Ages[Stage];
  if(!next || Age==next && ++Stage == 5)
  { Flag.bDead=true;
    g_World.DestroyMap(this);
  }
}


/* ObjFactory */
ObjFactory::ObjFactory()
{ assert(OT_NUM==1);
  Objs[OT_DEADSHIP] = NEW DeadShipObj;
}

ObjFactory::~ObjFactory()
{ for(int i=0; i<sizeof(Objs)/sizeof(Object*); i++) Objs[i]->Destroy();
}

Object * ObjFactory::Create(MsgNewObject *msg)
{ ObjType type = msg->OType;
  assert(type>=0 && type<OT_NUM);
  return Objs[type]->Clone(msg);
}


/* statics */
BlkAlloc<SputterObj> SputterObj::Alloc(128);
BlkAlloc<ExplodeObj> ExplodeObj::Alloc(32);

ExplodeObj::Explosion ExplodeObj::Types[3] =
{ { { SPR_EXP1_1, SPR_EXP1_2, SPR_NULL,   SPR_NULL,   SPR_NULL   },
    { 6, 12, 0, 0, 0 },
    3
  },
  { { SPR_EXP2_1, SPR_EXP2_2, SPR_EXP2_3, SPR_NULL,   SPR_NULL   },
    { 6, 12, 24, 0, 0 },
    4
  },
  { { SPR_EXP3_1, SPR_EXP3_2, SPR_EXP3_3, SPR_EXP3_4, SPR_EXP3_5 },
    { 6, 12, 20, 32, 50 },
    5
  }
};
