#ifndef GAMEOBJ_H_INC
#define GAMEOBJ_H_INC

enum ObjType
{ OT_DEADSHIP, OT_NUM
};

#ifndef TCSERVER

#include "Utility\Vector.h"
#include "Utility\BlkAlloc.h"
#include "World.h"

enum ObjBoundType
{ OBJ_UNKNOWN, OBJ_POINT, OBJ_SPHERE
};

class  ShipObj;
struct MsgNewObject;

class Object
{ public:
    Object();

    virtual void Draw(int x, int y)=0;
    virtual void CalcVel();
    virtual bool IsHitMap()=0;
    virtual void HitMap()             { Flag.bDead=true; }
    virtual void HitShip(ShipObj *)   { Flag.bDead=true; }
    virtual void HitObject(Object *)  { }
    virtual void Damage(ShipObj *s)   { }
    virtual bool Intersects(Object *)=0;
    virtual void Tick(unsigned)       { }
    virtual Object * Clone(MsgNewObject*)  { assert(0); return NULL; }
    virtual void Destroy()=0;

    void AddForce(FLOAT xv, FLOAT yv)         { Vel.X+=xv, Vel.Y+=yv;  }
    void AddForce(const GameLib::Vector<FLOAT> &force) { Vel+=force;   }
    void Move()                                        { Pos+=Vel;     }
    void Rotate(int angle)                             { Angle=(byte)(Angle+angle); }

    union
    { struct
      { unsigned bDead   :1;
        unsigned bNoGrav :1;
        unsigned bHitObjs:1;
        unsigned bRest   :1;
        unsigned bOnBase :1;
        unsigned bAccel  :1;
        unsigned bHidden :1;
      } Flag;
      byte Flags;
    };

    GameLib::Vector<FLOAT> Pos, Vel;
    int       Angle, Weight, Index, BoundType;
    unsigned  ID, Owner, Age;
    byte      Color;
};

class PointObj : public Object
{ public:
    PointObj() { BoundType=OBJ_POINT; }
    PointObj(int x, int y, FLOAT xv=0, FLOAT yv=0) { BoundType=OBJ_POINT; }

    virtual bool Intersects(Object *);
    virtual bool IsHitMap()
    { return g_World.IsHitMap((int)Pos.X, (int)Pos.Y) != 0; }
};

class SphereObj : public PointObj
{ public:
    SphereObj() { Radius=0; BoundType=OBJ_SPHERE; }
    SphereObj(int x, int y, int rad=0, FLOAT xv=0, FLOAT yv=0)
      : PointObj(x, y, xv, yv) { Radius=rad; BoundType=OBJ_SPHERE; }

    virtual bool Intersects(Object *);
    virtual bool IsHitMap()
    { return g_World.IsHitMap((int)Pos.X, (int)Pos.Y) != 0; }

    int Radius;
};


class NullPtObj : public PointObj
{ void Draw(int, int) { }
};

class SputterObj : public PointObj
{ public:
    virtual void Draw(int, int);
    virtual void Tick(unsigned) { if(Age>Life) Flag.bDead=true; }
    virtual void Destroy()      { Alloc.Delete(this); }
    int Life;

    static GameLib::BlkAlloc<SputterObj> Alloc;
};

class ExplodeObj : public SphereObj
{ public:
    ExplodeObj();
    virtual void Draw(int, int);
    virtual void Tick(unsigned);
    virtual void HitMap()       { g_World.DestroyMap(this); Vel=GameLib::Vector<FLOAT>(); }
    virtual void HitShip(ShipObj *);
    virtual void Destroy()      { Alloc.Delete(this); }
    int  Type, Stage, Activate, HitCounter;

    static GameLib::BlkAlloc<ExplodeObj> Alloc;
    static struct Explosion
    { RECT Rects[5];
      int  Ages [5];
      int  Radius;
    } Types[3];
};

struct MsgNewObject;
class ObjFactory
{ public:
    ObjFactory();
   ~ObjFactory();

    Object * Create(MsgNewObject *);
  
  protected:
    Object *Objs[OT_NUM];
};

#endif // ndef TCSERVER

#endif