#include "..\Precomp.h"
#include "Ship.h"
#include "IntrFace\Prims2D.h"
#include "Utility\GenMath.h"
#include "Utility\BlkAlloc.h"
#include "..\Network\Network.h"

/* FlameObj */
class FlameObj : public SputterObj
{ public:
    virtual void Tick(unsigned);
    static BlkAlloc<FlameObj> Alloc;
};
BlkAlloc<FlameObj> FlameObj::Alloc(128);

void FlameObj::Tick(unsigned)
{ if(!--Life) Flag.bDead=true;
  else if(Color==RC_LTGREY && Life<7) Color=RC_DKGREY;
  else if(Color==RC_DKBLUE && Life<9) Color=RC_YELLOW;
  else if(Color==RC_YELLOW && Life<6) Color=RC_ORANGE;
  else if(Color==RC_ORANGE && Life<4) Color=RC_RED;
}


/* ShipObj */
ShipObj::ShipObj()
{ Radius = 9;
  Gun    = Special  = NULL;
  nGun   = nSpecial = WT_NONE;
  Color  = RC_LTGREEN;
  Init();
}

ShipObj::~ShipObj()
{ delete Gun;
  delete Special;
}

void ShipObj::Draw(int x, int y)
{ Vector<FLOAT> vL=Vector<FLOAT>(0, -SHIPSIZE), v1, v2=g_Vects[(byte)Angle]*SHIPSIZE,
                v3=g_Vects[(byte)(140*DEGSTO256+Angle)]*SHIPSIZE;
  int c, c2;
  if(Flash) c=c2=RC_WHITE, Flash--;
  else c=Color, c2=c-2;

  if(c==c2)
  { Tri2D(g_Back, MakeVertex(v1,x,y), MakeVertex(v2,x,y),
          MakeVertex(v3,x,y), c, ST_FLAT);
    v3 = g_Vects[(byte)(220*DEGSTO256+Angle)]*SHIPSIZE;
    Tri2D(g_Back, MakeVertex(v1,x,y), MakeVertex(v2,x,y),
          MakeVertex(v3,x,y), c, ST_FLAT);
  }
  else
  { Tri2D(g_Back, MakeVertex(v1,x,y,c), MakeVertex(v2,x,y,c2),
          MakeVertex(v3,x,y,c2), ST_GOURAUD);
    v3 = g_Vects[(byte)(220*DEGSTO256+Angle)]*SHIPSIZE;
    Tri2D(g_Back, MakeVertex(v1,x,y,c), MakeVertex(v2,x,y,c2),
          MakeVertex(v3,x,y,c2), ST_GOURAUD);
  }
}

void ShipObj::HitMap()
{ Vector<FLOAT> vel = Vel;
  Impact(this, 0, 6, 5, false);
  Pos = OldPos;
  Vel.X = -Vel.X/5, Vel.Y = -Vel.Y/5;
  if(fabs(Vel.X) < 0.1) Vel.X = 0;
  if(fabs(Vel.Y) < 0.1) Vel.Y = 0;
}

void ShipObj::HitShip(ShipObj *s)
{ return;
  /*Vector<FLOAT> vel, tmp=Vel;
  Impact(s, 0, 6, 5);
  vel = Vel, Vel = tmp;
  s->Impact(this, 0, 6, 5);
  Vel = vel, Pos = OldPos, s->Pos = s->OldPos;
  Vel.X = -Vel.X/5, Vel.Y = -Vel.Y/5;
  if(fabs(Vel.X) < 0.1) Vel.X = 0;
  if(fabs(Vel.Y) < 0.1) Vel.Y = 0;

  s->Vel.X = -s->Vel.X/5, s->Vel.Y = -s->Vel.Y/5;
  if(fabs(s->Vel.X) < 0.1) s->Vel.X = 0;
  if(fabs(s->Vel.Y) < 0.1) s->Vel.Y = 0;*/
}

void ShipObj::Tick(unsigned tick)
{ if(Gun)     Gun->Tick(tick);
  if(Special) Special->Tick(tick);
  if(Flag.bOnBase)
  { if(Flag.bRest)
    { if((signed char)Angle<0) Angle=(byte)(Angle+1);
      else if((signed char)Angle>0) Angle=(byte)(Angle-1);
    }
    if(&g_World.Me->Ship==this)
    { if((tick&1) && ++Health > MaxHealth) Health=MaxHealth;
      if(Gun && Gun->Ammo<Gun->MaxAmmo && !(tick%Gun->Fill)) Gun->Ammo++;
      if(Special && Special->Ammo<Special->MaxAmmo && !(tick%Special->Fill))
        Special->Ammo++;
      if((Fuel += 4) > MaxFuel)  Fuel=MaxFuel;
    }
    else
    { Fuel   = MaxFuel;
      Health = MaxHealth;
      if(Gun)     Gun->Ammo = Gun->MaxAmmo;
      if(Special) Special->Ammo = Special->MaxAmmo;
    }
  }
}

void ShipObj::Init()
{ Fuel   = MaxFuel   = 3000;
  Health = MaxHealth = 250;
  Weight = 50;
  Flash  = TurnAcc = 0;
  bAlive = true;
}

void ShipObj::Accel()
{ if(Fuel) Fuel--;
  else if(Health) Health--;
  else return;
  AddFlame();
  AddForce(g_Vects[(byte)Angle]*0.042);
  Flag.bRest = Flag.bOnBase = false;
}

void ShipObj::Impact(Object *o, int extra, int mul, int div, bool notland)
{ assert(o->Weight);
  Vector<FLOAT> vel=o->Vel, force=vel;
  int dam;

  if(o != this)
  { if(notland) force = force*o->Weight/Weight;
    vel = (Vel-vel)*o->Weight/Weight;
  }

  if(div) vel = vel*mul/div;
  if(dam = Round(vel.Length_Sqr())+extra)
  { Flash = 4;
    Flag.bRest = Flag.bOnBase = false;
  }
  else
  { if(!notland)
    { if(!Flag.bRest)
      { Flag.bRest=true;
        TurnAcc=1;
      }
      if(!Flag.bOnBase && IsOnBase()) Flag.bOnBase=true;
    }
  }
  Health -= dam;
  if(notland) AddForce(force);
}

void ShipObj::Shoot()
{ if(Gun && Gun->bFiring) Gun->Fire();
  if(Special && Special->bFiring) Special->Fire();
}

void ShipObj::CtlFire(GunPos gun, bool on)
{ WepBase *wep = (gun==GP_GUN ? Gun : Special);
  if(wep) wep->CtlFire(on);
}

void ShipObj::SetWeps(WepType gun, WepType spec)
{ if(gun != WT_NONE && nGun != gun)
  { if(Gun) Gun->Destroy();
    Gun = g_WepFact.Create(gun, this);
    nGun = gun;
  }
  if(spec != WT_NONE && nSpecial != spec)
  { if(Special) Special->Destroy();
    Special = g_WepFact.Create(spec, this);
    nSpecial = spec;
  }
}

void ShipObj::AddFlame()
{ SputterObj *o = FlameObj::Alloc.New();
  FLOAT scale = Vel.Length_Sqr()/10;
  if(scale>1.8) scale=1.8;

  o->Pos   = Pos-g_Vects[Angle]*(SHIPSIZE/2);
  o->Vel   = Vel-g_Vects[(byte)(Angle+rand()%29-14)]*scale;
  o->Color = (Health < (MaxHealth>>2) ? RC_LTGREY : RC_DKBLUE);
  o->Life  = 9+rand()%4;
  g_World.AddObject(o);
}

GunPos ShipObj::FindWep(WepType wt)
{ if(Gun && Gun->Type==wt) return GP_GUN;
  if(Special && Special->Type==wt) return GP_SPECIAL;
  assert(0);
  return GP_UNKNOWN;
}


/* DeadShipObj */
Object * DeadShipObj::Clone(MsgNewObject *msg)
{ DStruct     *d = (DStruct*)msg->ObjData;
  DeadShipObj *s = NEW DeadShipObj;
  s->LoadShip(&g_World.Players[g_World.FindPlayer(msg->ID)]->Ship);
  s->Health = d->Health, s->Owner = msg->ID, s->ID = msg->ObjID;
  return s;
}

void DeadShipObj::LoadShip(ShipObj *s)
{ *((ShipObj*)this)=*s;
  Gun   = Special = NULL;
  Color = RC_WHITE;
  ID    = 0;
  Flags = 0;
  Flag.bHitObjs = true;
}

void DeadShipObj::Tick(unsigned tick)
{ if(Health<=-50)
  { MsgRemObject msg;
    msg.ObjID = ID;
    g_World.Send(&msg, NET_SENDGUAR);
    g_World.MkExp(ET_MEDIUM, (int)Pos.X, (int)Pos.Y, &g_World.Me->Ship);
    Flag.bDead = true;
  }
  else if(!(tick&3)) OldPos=Pos;
}
