#include "..\Precomp.h"
#include "Weapons.h"
#include "Ammo.h"
#include "..\Network\Network.h"

/* Global */
WepFactory g_WepFact;
const char *g_WepName[WT_NUM] =
{ // guns
  "Single", "Dual", "Two-Way", "Wavy",
  // specials
  "Cannon", "Grenade", "Drop Bomb", "Dynamite"
};

inline static AmmoBullet * NewBullet()
{ return (AmmoBullet*)g_AmmoFact.Create(AT_BULLET);
}

/* WepBase */
void WepBase::CtlFire(bool on)
{ bFiring = on;
  if(g_World.bNet && Ship == &g_World.Me->Ship && bFast)
  { if(on)
    { MsgWeaponStart start;
      start.WType = Type;
      start.ID    = Ship->ID;
      g_World.Send(&start, NET_SENDGUAR);
    }
    else
    { MsgWeaponEnd end;
      end.WType = Type;
      end.ID    = Ship->ID;
      g_World.Send(&end, NET_SENDGUAR);
    }
  }
}

void WepBase::NetFire(AmmoPoint *a, AmmoType type, void *data, int datalen)
{ assert(g_World.bNet);
  PointObj      *obj = a;
  MsgWeaponFire *msg = (MsgWeaponFire *)_alloca(sizeof(MsgWeaponFire)+datalen);
  place_new(msg) MsgWeaponFire;
  msg->AType = type;
  msg->X  = obj->Pos.X, msg->Y  = obj->Pos.Y;
  msg->XV = obj->Vel.X, msg->YV = obj->Vel.Y;
  msg->ID = Ship->ID,   msg->Color = obj->Color;
  msg->Weight = (WORD)obj->Weight, msg->Extra = (WORD)a->Extra;
  if(datalen)
  { memcpy(&msg->AmmoData, data, datalen);
    msg->Len += datalen;
  }
  g_World.Send(msg, NET_SENDGUAR);
}

void WepBase::NetFire(AmmoSphere *a, AmmoType type, void *data, int datalen)
{ assert(g_World.bNet);
  SphereObj     *obj = a;
  MsgWeaponFire *msg = (MsgWeaponFire *)_alloca(sizeof(MsgWeaponFire)+datalen);
  place_new(msg) MsgWeaponFire;
  msg->AType = type;
  msg->X  = obj->Pos.X, msg->Y  = obj->Pos.Y;
  msg->XV = obj->Vel.X, msg->YV = obj->Vel.Y;
  msg->ID = Ship->ID,   msg->Radius = obj->Radius;
  msg->Weight = (WORD)obj->Weight, msg->Extra = (WORD)a->Extra;
  msg->Color  = obj->Color;
  if(datalen)
  { memcpy(&msg->AmmoData, data, datalen);
    msg->Len += datalen;
  }
  g_World.Send(msg, NET_SENDGUAR);
}

/* WepMachineGun */
WepMachineGun::WepMachineGun(ShipObj *s)
  : WepBase(s)
{ MaxAmmo  = Ammo = 400;
  Fill     = 2;
  bFast    = true;
  Type     = WT_MACHINEGUN;
}

void WepMachineGun::Fire()
{ if(!Reload && Ammo)
  { AmmoBullet *bullet = NewBullet();
    Vector<FLOAT> vel = g_Vects[Ship->Angle]*3;
    bullet->Pos = Ship->Pos+vel;
    bullet->Vel = Ship->Vel+vel;
    bullet->LoadShip(Ship);
    g_World.AddObject(bullet);
    Ammo--;
    Reload = 4;
  }
}


/* WepDualMachGun */
WepDualMachGun::WepDualMachGun(ShipObj *s)
  : WepBase(s)
{ MaxAmmo = Ammo = 400;
  Fill    = 2;
  bFast   = true;
  Type    = WT_DUAL_MACHINEGUN;
}

void WepDualMachGun::Fire()
{ if(!Reload && Ammo)
  { AmmoBullet *b1 = NewBullet();
    AmmoBullet *b2 = NewBullet();
    b1->Vel = b2->Vel = g_Vects[Ship->Angle]*3+Ship->Vel;
    b1->Pos = Ship->Pos+Vector<FLOAT>(-3, -4).RotatedZ(Ship->Angle);
    b2->Pos = Ship->Pos+Vector<FLOAT>( 3, -4).RotatedZ(Ship->Angle);
    b1->LoadShip(Ship);
    b2->LoadShip(Ship);
    g_World.AddObject(b1);
    g_World.AddObject(b2);
    if((Ammo-=2) < 0) Ammo=0;
    Reload = 6;
  }
}


/* WepFBMachGun */
WepFBMachGun::WepFBMachGun(ShipObj *s)
  : WepBase(s)
{ MaxAmmo = Ammo = 420;
  Fill    = 2;
  bFast   = true;
  Type    = WT_FB_MACHINEGUN;
}

void WepFBMachGun::Fire()
{ if(!Reload && Ammo)
  { AmmoBullet *b1 = NewBullet();
    AmmoBullet *b2 = NewBullet();
    Vector<FLOAT> vel = g_Vects[Ship->Angle]*3;
    b1->Pos = Ship->Pos+vel, b2->Pos = Ship->Pos-vel;
    b1->Vel = Ship->Vel+vel, b2->Vel = Ship->Vel-vel;
    b1->LoadShip(Ship);
    b2->LoadShip(Ship);
    g_World.AddObject(b1);
    g_World.AddObject(b2);
    if((Ammo-=2) < 0) Ammo=0;
    Reload = 4;
  }
}


/* WepWavyMachGun */
WepWavyMachGun::WepWavyMachGun(ShipObj *s)
  : WepBase(s)
{ MaxAmmo = Ammo = 400;
  Fill    = 2;
  bFast   = true;
  Type    = WT_WAVY_MACHINEGUN;
}

void WepWavyMachGun::Fire()
{ if(!Reload && Ammo)
  { AmmoBullet *b = NewBullet();
    b->Pos = Ship->Pos + Vector<FLOAT>(4*SineTable.Sin(TCount<<4), -4).RotatedZ(Ship->Angle);
    b->Vel = Ship->Vel + g_Vects[Ship->Angle]*3;
    b->LoadShip(Ship);
    g_World.AddObject(b);
    Ammo--;
    Reload = 4;
  }
}


/* WepCannon */
WepCannon::WepCannon(ShipObj *s)
  : WepBase(s)
{ MaxAmmo = Ammo = 25;
  Fill    = 20;
  bFast   = false;
  Type    = WT_CANNON;
}

void WepCannon::Fire()
{ if(!Reload && Ammo)
  { AmmoCannonBall *b = (AmmoCannonBall*)(g_World.bNet ?
                          g_AmmoFact.Static(AT_CANNONBALL) :
                          g_AmmoFact.Create(AT_CANNONBALL));
    b->Pos = Ship->Pos+g_Vects[Ship->Angle]*2;
    b->Vel = Ship->Vel+g_Vects[Ship->Angle]*3;
    b->LoadShip(Ship);
    if(g_World.bNet) NetFire(b, AT_CANNONBALL);
    else g_World.AddObject(b);
    
    Ammo--;
    Reload = 60;
    // recoil
    Ship->Vel -= g_Vects[Ship->Angle]*0.40;
  }
}


/* WepGrenader */
WepGrenader::WepGrenader(ShipObj *s)
  : WepBase(s)
{ MaxAmmo = Ammo = 20;
  Fill    = 20;
  bFast   = false;
  Type    = WT_GRENADER;
}

void WepGrenader::Fire()
{ if(!Reload && Ammo)
  { AmmoGrenade *g = (AmmoGrenade*)(g_World.bNet ?
                        g_AmmoFact.Static(AT_GRENADE) :
                        g_AmmoFact.Create(AT_GRENADE));
    g->Pos = Ship->Pos+g_Vects[Ship->Angle]*2;
    g->Vel = Ship->Vel+g_Vects[Ship->Angle]*2.5;
    g->LoadShip(Ship);
    if(g_World.bNet) NetFire(g, AT_GRENADE);
    else g_World.AddObject(g);

    Ammo--;
    Reload = 60;
  }
}


/* WepDropBomber */
WepDropBomber::WepDropBomber(ShipObj *s)
  : WepBase(s)
{ MaxAmmo = Ammo = 60;
  Fill    = 9;
  bFast   = true;
  Type    = WT_DROPBOMBER;
}

void WepDropBomber::Fire()
{ if(!Reload && Ammo)
  { AmmoDropBomb *b = (AmmoDropBomb *)g_AmmoFact.Create(AT_DROPBOMB);
    Vector<FLOAT> vec = g_Vects[Ship->Angle]*2;
    b->Pos = Ship->Pos - vec, b->Vel = Ship->Vel - vec;
    b->LoadShip(Ship);
    g_World.AddObject(b);
    Ammo--;
    Reload = 12;
  }
}


/* WepDynamiter */
WepDynamiter::WepDynamiter(ShipObj *s)
  : WepBase(s)
{ MaxAmmo = Ammo = 14;
  Fill    = 24;
  bFast   = false;
  Type    = WT_DYNAMITER;
}

void WepDynamiter::Fire()
{ if(!Reload && Ammo)
  { AmmoDynamite *d = (AmmoDynamite *)(g_World.bNet ?
                        g_AmmoFact.Static(AT_DYNAMITE) :
                        g_AmmoFact.Create(AT_DYNAMITE));
    d->Pos = Ship->Pos-g_Vects[Ship->Angle]*SHIPSIZE;
    d->LoadShip(Ship);
    if(g_World.bNet) NetFire(d, AT_DYNAMITE);
    else g_World.AddObject(d);
    Ammo--;
    Reload = 80;
  }
}


/* WepFactory */
WepFactory::WepFactory()
{ assert(WT_NUM==8);
  Weps[WT_MACHINEGUN]      = NEW WepMachineGun;
  Weps[WT_DUAL_MACHINEGUN] = NEW WepDualMachGun;
  Weps[WT_FB_MACHINEGUN]   = NEW WepFBMachGun;
  Weps[WT_WAVY_MACHINEGUN] = NEW WepWavyMachGun;
  Weps[WT_CANNON]          = NEW WepCannon;
  Weps[WT_GRENADER]        = NEW WepGrenader;
  Weps[WT_DROPBOMBER]      = NEW WepDropBomber;
  Weps[WT_DYNAMITER]       = NEW WepDynamiter;
}

WepFactory::~WepFactory()
{ for(int i=0; i<sizeof(Weps)/sizeof(WepBase*); i++) Weps[i]->Destroy();
}

WepBase * WepFactory::Create(WepType type, ShipObj *s)
{ assert(type != WT_NONE);
  return Weps[type]->Clone(s);
}
