#ifndef WEAPONS_H_INC
#define WEAPONS_H_INC

enum WepType
{ // guns
  WT_MACHINEGUN, WT_DUAL_MACHINEGUN, WT_FB_MACHINEGUN,
  WT_WAVY_MACHINEGUN,
  WT_GUNMAX=WT_WAVY_MACHINEGUN, 
  // specials
  WT_CANNON, WT_GRENADER, WT_DROPBOMBER, WT_DYNAMITER,
  WT_SPECMAX=WT_DYNAMITER,
  // etc
  WT_NUM, WT_NONE,
  WT_DEFGUN=WT_MACHINEGUN, WT_DEFSPEC=WT_CANNON // defaults
};

#ifndef TCSERVER

extern const char *g_WepName[WT_NUM];

#include "GameObj.h"

class AmmoPoint;
class AmmoSphere;

class WepBase
{ public:
    WepBase() { };
    WepBase(ShipObj *s) { Ship=s; Reload=0; bFiring=false; }

    virtual void Destroy() { delete this; }
    virtual void Fire()=0;
    virtual void Tick(unsigned) { if(Reload) Reload--; }
    virtual WepBase * Clone(ShipObj *)=0;
    void CtlFire(bool on);
    void NetFire(AmmoPoint  *, AmmoType type, void *data=NULL, int datalen=0);
    void NetFire(AmmoSphere *, AmmoType type, void *data=NULL, int datalen=0);

    ShipObj *Ship;
    int      Ammo, MaxAmmo, Reload, Fill;
    WepType  Type;
    bool     bFiring, bFast;
};

class WepMachineGun : public WepBase
{ public:
    WepMachineGun() { };
    WepMachineGun(ShipObj *s);
    virtual void Fire();
    virtual WepBase *Clone(ShipObj *s) { return NEW WepMachineGun(s); }
};

class WepDualMachGun : public WepBase
{ public:
    WepDualMachGun() { };
    WepDualMachGun(ShipObj *s);
    virtual void Fire();
    virtual WepBase *Clone(ShipObj *s) { return NEW WepDualMachGun(s); }
};

class WepFBMachGun : public WepBase
{ public:
    WepFBMachGun() { };
    WepFBMachGun(ShipObj *s);
    virtual void Fire();
    virtual WepBase *Clone(ShipObj *s) { return NEW WepFBMachGun(s); }
};

class WepWavyMachGun : public WepBase
{ public:
    WepWavyMachGun() { };
    WepWavyMachGun(ShipObj *s);
    virtual void Fire();
    virtual void Tick(unsigned t) { if(Reload) Reload--; TCount=t; }
    virtual WepBase *Clone(ShipObj *s) { return NEW WepWavyMachGun(s); }

    int TCount;
};

class WepCannon : public WepBase
{ public:
    WepCannon() { };
    WepCannon(ShipObj *s);
    virtual void Fire();
    virtual WepBase *Clone(ShipObj *s) { return NEW WepCannon(s); }
};

class WepGrenader : public WepBase
{ public:
    WepGrenader() { };
    WepGrenader(ShipObj *s);
    virtual void Fire();
    virtual WepBase *Clone(ShipObj *s) { return NEW WepGrenader(s); }
};

class WepDropBomber : public WepBase
{ public:
    WepDropBomber() { };
    WepDropBomber(ShipObj *s);
    virtual void Fire();
    virtual WepBase *Clone(ShipObj *s) { return NEW WepDropBomber(s); }
};

class WepDynamiter : public WepBase
{ public:
    WepDynamiter() { };
    WepDynamiter(ShipObj *s);
    virtual void Fire();
    virtual WepBase *Clone(ShipObj *s) { return NEW WepDynamiter(s); }
};

class WepFactory
{ public:
    WepFactory();
   ~WepFactory();

    WepBase * Create(WepType type, ShipObj *s);

  protected:
    WepBase *Weps[WT_NUM];
};

#endif // ndef TCSERVER

#endif
