#ifndef WORLD_H_INC
#define WORLD_H_INC

enum ExpType
{ ET_TINY, ET_SMALL, ET_MEDIUM, ET_LARGE, ET_HUGE, ET_ARMAGEDDON, ET_NUM
};

#ifndef TCSERVER

class Object;
class Player;
class ExplodeObj;
class ShipObj;
struct NetMsg;
struct CircleRem;

namespace GameLib
{ class CDDSurface;
}

class CWorld
{ public:
    CWorld();
   ~CWorld();

    int  InitNet();
    int  Load(const char *fore, const char *back=NULL);
    int  LoadNet(const char *ip, const char *name);
    void Unload();

    void SetTiming();
    void DoTicks();
    void Advance();
    void Render();

    int  Connect(const char *ip);
    void Disconnect();
    int  Send(NetMsg *, DWORD flags=0, DWORD msTimeout=0, DPNHANDLE *hand=NULL);

    void MkExp(ExpType, int x, int y, ShipObj *, int seed=0);
    void Explosion(ExpType type, int x, int y, ShipObj *);
    void DestroyMap(ExplodeObj *);
    void AddExp(ExplodeObj *);

    int IsHitMap(int x, int y)
    { assert(bLocked);
      if((unsigned)x>Width || (unsigned)y>Height) return 1;
      register byte b=Map->Mem[y*Map->Pitch+x];
      return C_HIT(b);
    }
    int IsOnBase(int x, int y)
    { assert(bLocked);
      if((unsigned)x>Width || (unsigned)y>Height) return 0;
      register byte b=Map->Mem[y*Map->Pitch+x];
      return C_BASE(b);
    }

    void AddObject(Object *o);
    void RemObject(Object *o);
    int  FindObject(unsigned id);

    void AddPlayer(Player *p);
    void RemPlayer(Player *p);
    int  FindPlayer(unsigned id);

    void SetMe(Player *p)     { Me=p;     }
    void SetCamera(Object *o) { Camera=o; }

    void DoInput();

    void Lock()   { EnterCriticalSection(&CS); }
    void Unlock() { LeaveCriticalSection(&CS); }

    void AddLine(const char *, ...);

    CRITICAL_SECTION      CS;
    GameLib::CDDSurface  *Map, *Back;
    Player    **Players, *Me;
    Object    **Objects, *Camera;

    CircleRem  *Circs;
    int         nCircs, maxCircs;
    int         nGun, nSpecial;
    __int64     tCur, tNext, tInc;
    FLOAT       Gravity;
    std::string Name;
    int         nObjects, maxObjects, nPlayers, maxPlayers;
    unsigned    Width, Height, Tick, State, OldState;
    DPNHANDLE   hMoveMsg;
    byte        Reason;
    bool        bLocked, bNet, bQuit, bMenuGun;
    byte        Palette[768];

    void        MsgHandler(NetMsg *);

  protected:
    struct Line_s
    { ~Line_s() { }
      std::string  Line;
      int          Age, MaxAge;
    }  *Lines;
    int nLines, maxLines;

    int  LoadLevels(const char *fore, const char *back);
    void AdvanceWorld();
    void RenderWorld();
    void SendMove();
    void SetState(int newstate) { OldState=State, State=newstate; }
    bool RemoveCircle(int x, int y, int rad);
    bool FlatRegion(int y, int start, int end);
    bool FlatSpan8C(byte *mem, int start, int end, int wid);

    IDirectPlay8Client *pDPC;
    static HRESULT WINAPI Handler(PVOID, DWORD, PVOID);
};

#endif // ndef TCSERVER

#endif
