#include "..\Precomp.h"
#include "Systems.h"
#include <dinput.h>
/*** GLOBAL VARS ***/
CSystems    g_Sys;
CDInput     g_Input;
CDIKeyboard g_KB;

/*** CSystems IMPLEMENTATION ***/
CSystems::CSystems()
{ LastError = SYS_OK;
  restMode  = bGraph = bInput = false;
}

CSystems::~CSystems()
{ DeinitGraphics();
  DeinitInput();
}

SYSErrors CSystems::InitGraphics()
{ assert(g_Win.hWnd);
  DeinitGraphics();
  g_Win.Focus();
  if(g_DD.Init(g_Win.hWnd)) return SYSError(SYS_DDINIT);
  
  if(g_Win.bFull)
  { CDDDepth dep = CDD_8BIT;
    if(g_DD.SetMode(g_Win.Width, g_Win.Height, CDD_8BIT, CDD_FULLSCREEN, g_Win.bStdVGA))
      SYSError(SYS_DDSETMODE);
  }
  else
  { DEVMODE dm;
    // get the current display settings and try to change to 16, 24, or 32
    if(!::EnumDisplaySettings(NULL, ENUM_CURRENT_SETTINGS, &dm))
      return SYSError(SYS_DDCHGMODE);
    if(dm.dmBitsPerPel != 8)
    { dm.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
      devMode = dm;
      dm.dmBitsPerPel = 8;
      if(::ChangeDisplaySettings(&dm, 0) != DISP_CHANGE_SUCCESSFUL)
        return SYSError(SYS_DDCHGMODE);
      restMode = true;
    }
    // display mode was set.. now try to initialize CDDraw
    if(g_DD.SetMode(g_Win.Width, g_Win.Height, CDD_8BIT, CDD_WINDOWED))
      return SYSError(SYS_DDSETMODE);
  }

  // create a single buffered display with a back buffer in system memory
  if(g_DD.CreatePrimary(CDD_SINGLE)) return SYSError(SYS_DDCREATEPRIM);
  g_Back = NEW CDDSurface;
  if(g_Back->Init(&g_DD, g_DD.Primary.Width, g_DD.Primary.Height, CDD_SYSTEM))
    return SYSError(SYS_DDINITBACK);

  if(!g_Sprites)
  { g_Sprites = g_VFS.LoadPcx("sprites.pcx", NULL, CDD_SYSTEM);
    if(!g_Sprites) return SYSError(SYS_DDINITBACK);
    g_Sprites->Stats.Trans=true;
  }

  bGraph = true;
  return SYS_OK;
}

void CSystems::DeinitGraphics()
{ bGraph = false;
  if(g_Sprites)
  { g_Sprites->Deinit();
    delete g_Sprites;
    g_Sprites = NULL;
  }
  if(g_Back)
  { g_Back->Deinit();
    delete g_Back;
    g_Back = NULL;
  }
  g_DD.Deinit();
  if(restMode)
  { ::ChangeDisplaySettings(&devMode, 0);
    restMode=false;
  }
}

SYSErrors CSystems::InitInput()
{ g_KB.SetCoopLevel(CDI_EXCLUSIVE);
  if(g_Input.Init(g_Win.hWnd, g_Win.hInst)) return SYSError(SYS_DIINIT);
  if(g_Input.RegDevice(&g_KB)) return SYSError(SYS_DIREGKB);
  if(g_Input.Start()) return SYSError(SYS_DISTART);
  g_KB.Acquire();
  bInput = true;
  return SYS_OK;
}

void CSystems::DeinitInput()
{ bInput = false;
  g_KB.Unacquire();
  g_Input.Stop();
  g_Input.UnregDevice(&g_KB);
  g_Input.Deinit();
}

void CSystems::FlipScreen()
{ assert(bGraph);
  g_Back->Flip();
}
