#ifndef SYSTEMS_H_INC
#define SYSTEMS_H_INC

enum SYSErrors // errors returned by the CSystems object
{ SYS_OK, SYS_DDINIT, SYS_DDSETMODE, SYS_DDCHGMODE, SYS_DDCREATEPRIM,
  SYS_DDINITBACK, SYS_DIINIT, SYS_DIREGKB, SYS_DISTART
};

namespace GameLib
{ class CDIDevice;
};
typedef void (*DICALLBACK)(GameLib::CDIDevice *);

class CSystems
{ public:
    CSystems();
    ~CSystems();

    // clip mouse coordinates to the screen
    static inline void ClipMouse(int *x, int *y)
    { if(*x<0) *x = 0;
      else if(*x>=g_Win.Width)  *x = g_Win.Width-1;
      if(*y<0) *y = 0;
      else if(*y>=g_Win.Height) *y = g_Win.Height-1;
    }

    SYSErrors InitGraphics();   // initialize the graphics system
    void      DeinitGraphics(); // deinitialize it
    SYSErrors InitInput();
    void      DeinitInput();
    void      FlipScreen();     // flip the back buffer to the screen

    bool      bGraph, bInput;   // are these systems initialized?

  protected:
    // return the index of this hotkey or -1 if not found
    SYSErrors SYSError(SYSErrors sye) { return LastError=sye; }
    SYSErrors LastError;     // the last error to occur

    DEVMODE   devMode;
    bool      restMode;

    friend class CWinApi;
};

#endif