#include "..\Server\TCServer.h"
#include "..\Precomp.h"
#include <direct.h>

/*** GLOBALS ***/
CWinApi g_Win;

/*** CWinApi IMPLEMENTATION ***/
CWinApi::CWinApi()
{ hWnd   = NULL;
  hInst  = NULL;
  bRun   = false;

#ifndef NDEBUG
  _CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_CHECK_ALWAYS_DF |
                 _CRTDBG_LEAK_CHECK_DF);
#endif

  // extract the path for the current executable
  int s=0, e=0, len = strlen(::GetCommandLineA());
  char *src, buf[_MAX_PATH];
  CmdLine = NEW char[len+1];
  strcpy(CmdLine, ::GetCommandLineA());

  while(s<len && isspace(CmdLine[s])) s++;
  if(s<len)
  { if(CmdLine[s] == '\"')
    { e=++s;
      while(e<len && CmdLine[e] != '\"') e++;
    }
    else
    { e=s;
      while(e<len && !isspace(CmdLine[e])) e++;
    }
    while(e>=s && CmdLine[e] != DIRSEPC) e--;
  }
  len = e-s+1;
  if(len<=1)
  { len=strlen(_getcwd(src=buf, _MAX_PATH));
    if(buf[len-1]!='\\') buf[len++]='\\';
  }
  else src=CmdLine+s;

  ExePath = NEW char[len+1];
  memcpy(ExePath, src, len);
  ExePath[len] = 0;
}

CWinApi::~CWinApi()
{ 
#ifndef TCSERVER
  DestroyWnd();
#endif
  delete[] ExePath;
  delete[] CmdLine;
}

#ifndef TCSERVER
int CWinApi::Init(HINSTANCE hinst)
{ WNDCLASS  wnd;  // register window class
  memset(&wnd, 0, sizeof(wnd));
  wnd.hInstance     = hInst = hinst;
  wnd.lpfnWndProc   = WndProc;
  wnd.lpszClassName = GAMECLASS;
  wnd.style         = CS_HREDRAW | CS_VREDRAW;
  return !::RegisterClass(&wnd);
}

int CWinApi::CreateWnd(int wid, int hei, bool fullscreen, bool stdvga)
{ assert(hInst);
  DestroyWnd();

  // different window styles for fullscreen or not
  DWORD style = fullscreen ? WS_POPUP :
                  WS_OVERLAPPEDWINDOW & ~WS_MAXIMIZEBOX;
  if(fullscreen) // fullscreen... create topmost and as big as possible
  { hWnd = ::CreateWindowEx(WS_EX_TOPMOST, GAMECLASS, GAMETITLE, style, 0, 0,
                            GetSystemMetrics(SM_CXSCREEN),
                            GetSystemMetrics(SM_CYSCREEN),
                            NULL, NULL, hInst, NULL);
  }
  else // windowed.. create the proper window
  { // move 50 pixels from top-left (yes, it's needed)
    RECT  rect = { 50, 50, wid+50, hei+50 };
    ::AdjustWindowRect(&rect, style, FALSE);
    hWnd = ::CreateWindow(GAMECLASS, GAMETITLE _T(" v. ") GAMEVER, style, rect.left,
                          rect.top, rect.right-rect.left, rect.bottom-rect.top,
                          NULL, NULL, hInst, NULL);
  }
  if(!hWnd) return 1;
  ShowWindow(hWnd, fullscreen ? SW_MAXIMIZE : SW_SHOW); // maximize or show
  UpdateWindow(hWnd); // repaint

  Width   = wid, Height = hei;  // save info
  bFull   = fullscreen;
  bRun    = true;
  bStdVGA = stdvga;
  return 0;
}

void CWinApi::DestroyWnd()
{ if(hWnd)  // destroy the window
  { ::DestroyWindow(hWnd);
    hWnd = NULL;
    bRun = false;
  }
}

bool CWinApi::Pump()  // pump waiting messages
{ MSG msg;
  while(::PeekMessage(&msg, hWnd, 0, 0, PM_REMOVE))
  { if(msg.message == WM_QUIT) return bRun=false; // false on quit
    ::TranslateMessage(&msg);
    ::DispatchMessage(&msg);
  }
  return bRun;  // otherwise the current running? value
}

bool CWinApi::WaitMsg()
{ MSG msg;
  if(::GetMessage(&msg, hWnd, 0, 0))  // same, but wait for a single message
  { ::TranslateMessage(&msg);
    ::DispatchMessage(&msg);
    return bRun;
  }
  else return bRun=false;
}

void CWinApi::Quit()  // quit gracefully
{ bRun = false;
}

void CWinApi::MsgBox(const char *str, ...) const
{ TCHAR buf[1024];
  va_list list;
  va_start(list, str);
  _vstprintf(buf, str, list);
  va_end(list);
  ::MessageBoxA(hWnd, buf, GAMETITLE, MB_OK);
}

void CWinApi::Raise(const char *file, int line, const TCHAR *s, ...) const
{ char buf[1024];
  va_list list;
  va_start(list, s);
  vsprintf(buf+sprintf(buf, "%s: line %d\n", file, line), s, list);
  va_end(list);
  ::MessageBoxA(hWnd, buf, GAMETITLE, MB_OK);
}

void CWinApi::DbgOut(const char *str, ...)
{ char buf[1024];
  va_list list;
  va_start(list, str);
  vsprintf(buf, str, list);
  va_end(list);
  ::OutputDebugStringA(buf);
}

/*** STATICS IMPLEMENTATIONS ***/
long CALLBACK CWinApi::WndProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{ //InputEvt inp;
  char buf[2]={0,0};
  switch(msg)
  { case WM_MOUSEMOVE:
      if(g_Win.bFull) ShowCursor(FALSE);  // hide cursor in fullscreen
      g_MouseX = LOWORD(lparam), g_MouseY = HIWORD(lparam);
      g_Sys.ClipMouse(&g_MouseX, &g_MouseY);
      /*inp.type = IT_MSMOVE;
      inp.data.md.x = g_MouseX, inp.data.md.y = g_MouseY;
      g_UIMan.DispatchInput(&inp);*/
      return 0;

    /*case WM_SYSKEYDOWN:
    case WM_KEYDOWN:
      if(wparam<256)
      { if(wparam == VK_CAPITAL)
        { SHORT s = GetKeyState(VK_CAPITAL);
          g_Keys[VK_CAPITAL] = (s&1) | (s>>8);
        }
        else if(lparam & (1<<24))
        { if(wparam == VK_SHIFT)        wparam = VK_RSHIFT;
          else if(wparam == VK_CONTROL) wparam = VK_RCONTROL;
          else if(wparam == VK_MENU)    wparam = VK_RMENU;
        }

        g_Keys[wparam] = 0x80;
        inp.data.kd.vkey=wparam;
        inp.data.kd.down = true;
        inp.data.kd.rep  = LOWORD(lparam);
        inp.type = IT_KB;
        g_UIMan.DispatchInput(&inp);
      }
      return 0;

    case WM_SYSKEYUP:
    case WM_KEYUP:
      if(wparam<256)
      { if(wparam == VK_CAPITAL)
        { SHORT s = GetKeyState(VK_CAPITAL);
          g_Keys[VK_CAPITAL] = (s&1) | (s>>8);
        }
        else if(lparam & (1<<24))
        { if(wparam == VK_SHIFT)        wparam = VK_RSHIFT;
          else if(wparam == VK_CONTROL) wparam = VK_RCONTROL;
          else if(wparam == VK_MENU)    wparam = VK_RMENU;
        }
        g_Keys[wparam] = 0;
        inp.data.kd.vkey=wparam;
        inp.data.kd.down = 0;
        inp.type = IT_KB;
        g_UIMan.DispatchInput(&inp);
      }
      return 0;
      
    case 0x020A: // WM_MOUSEWHEEL
      inp.type = IT_MSWHEEL;
      inp.data.md.btn = (short)HIWORD(wparam);
      inp.data.md.x = LOWORD(lparam), inp.data.md.y = HIWORD(lparam);
      g_Sys.ClipMouse(&inp.data.md.x, &inp.data.md.y);
      g_UIMan.DispatchInput(&inp);
      return 0;

    case WM_LBUTTONDOWN:
    case WM_RBUTTONDOWN:
    case WM_MBUTTONDOWN:
      inp.type = IT_MSCLICK;
      inp.data.md.btn  = (msg==WM_LBUTTONDOWN ? 0 : (msg==WM_RBUTTONDOWN?1:2));
      inp.data.md.down = true;
      inp.data.md.x = LOWORD(lparam), inp.data.md.y = HIWORD(lparam);
      g_Sys.ClipMouse(&inp.data.md.x, &inp.data.md.y);
      g_UIMan.DispatchInput(&inp);
      return 0;

    case WM_LBUTTONUP:
    case WM_RBUTTONUP:
    case WM_MBUTTONUP:
      inp.type = IT_MSCLICK;
      inp.data.md.btn  = (msg==WM_LBUTTONUP ? 0 : (msg==WM_RBUTTONUP?1:2));
      inp.data.md.down = false;
      inp.data.md.x = LOWORD(lparam), inp.data.md.y = HIWORD(lparam);
      g_Sys.ClipMouse(&inp.data.md.x, &inp.data.md.y);
      g_UIMan.DispatchInput(&inp);
      return 0;*/

    case WM_PAINT:
      // if repaint needed, repaint
      if(g_Sys.bGraph && !g_Win.bFull && !g_Win.bMinim) g_Sys.FlipScreen();
      break;

    case WM_ACTIVATE:
      // update focus and minimized flags
      g_Win.bFocus = LOWORD(wparam) != WA_INACTIVE;
      g_Win.bMinim = HIWORD(wparam) != 0;
      if(wparam != WA_INACTIVE && g_Sys.bGraph) g_DD.SetPal(g_World.Palette);
      if(g_Sys.bInput)
      { if(wparam != WA_INACTIVE) g_KB.Acquire();
        else g_KB.Unacquire();
      }
      return 0;

    case WM_SIZE:
    case WM_MOVE:
      // if windowed mode and graphics inited, resize viewport on move/size
      if(!g_Win.bFull && g_Sys.bGraph) g_DD.AutoViewport();
      return 0;

    case WM_CLOSE:
    case WM_DESTROY:
      ::PostQuitMessage(0); // quit
      return 0;
  }
  return DefWindowProc(hwnd, msg, wparam, lparam);
}
#endif
