#include "..\Precomp.h"
#include "Client.h"

const char * DPErrToString(HRESULT hRet)
{ switch(hRet)
  { case DPNSUCCESS_PENDING:
      return "An asynchronous operation has reached the point where it is successfully queued.";
    case S_OK:
      return "The operation completed successfully.";
    case DPNERR_ABORTED:
      return "The operation was canceled before it could be completed.";
    case DPNERR_ADDRESSING:
      return "The address specified is invalid.";
    case DPNERR_ALREADYCONNECTED:
      return "The object is already connected to the session.";
    case DPNERR_ALREADYCLOSING:
      return "An attempt to call the Close method on a session has been made more than once.";
    case DPNERR_ALREADYDISCONNECTING:
      return "The client is already disconnecting from the session.";
    case DPNERR_ALREADYINITIALIZED:
      return "The object has already been initialized.";
    case DPNERR_BUFFERTOOSMALL:
      return "The supplied buffer is not large enough to contain the requested data.";
    case DPNERR_CANNOTCANCEL:
      return "The operation could not be canceled.";
    case DPNERR_CANTCREATEGROUP:
      return "A new group cannot be created.";
    case DPNERR_CANTCREATEPLAYER:
      return "A new player cannot be created.";
    case DPNERR_CANTLAUNCHAPPLICATION:
      return "The lobby cannot launch the specified application.";
    case DPNERR_CONNECTING:
      return "The method is in the process of connecting to the network.";
    case DPNERR_CONNECTIONLOST:
      return "The service provider connection was reset while data was being sent.";
    case DPNERR_DATATOOLARGE:
      return "The application data is too large for the service provider's Maximum Transmission Unit.";
    case DPNERR_DOESNOTEXIST:
      return "Requested element is not part of the address.";
    case DPNERR_ENUMQUERYTOOLARGE:
      return "The query data specified is too large.";
    case DPNERR_ENUMRESPONSETOOLARGE:
      return "The response to an enumeration query is too large.";
    case DPNERR_EXCEPTION:
      return "An exception occurred when processing the request.";
    case DPNERR_GENERIC:
      return "An undefined error condition occurred.";
    case DPNERR_GROUPNOTEMPTY:
      return "The specified group is not empty.";
    case DPNERR_HOSTREJECTEDCONNECTION:
      return "The DPN_MSGID_INDICATE_CONNECT system message returned something other than S_OK in response to a connect request.";
    case DPNERR_HOSTTERMINATEDSESSION:
      return "The host in a peer session (with host migration enabled) terminated the session.";
    case DPNERR_INCOMPLETEADDRESS:
      return "The address specified is not complete.";
    case DPNERR_INVALIDADDRESSFORMAT:
      return "Address format is invalid.";
    case DPNERR_INVALIDAPPLICATION:
      return "The GUID supplied for the application is invalid.";
    case DPNERR_INVALIDCOMMAND:
      return "The command specified is invalid.";
    case DPNERR_INVALIDDEVICEADDRESS:
      return "The address for the local computer or adapter is invalid.";
    case DPNERR_INVALIDFLAGS:
      return "The flags passed to this method are invalid.";
    case DPNERR_INVALIDGROUP:
      return "The group ID is not recognized as a valid group ID for this game session.";
    case DPNERR_INVALIDHANDLE:
      return "The handle specified is invalid.";
    case DPNERR_INVALIDHOSTADDRESS:
      return "The specified remote address is invalid.";
    case DPNERR_INVALIDINSTANCE:
      return "The GUID for the application instance is invalid.";
    case DPNERR_INVALIDINTERFACE:
      return "The interface parameter is invalid. This value will be returned in a connect request if the connecting player was not a client in a client/server game or a peer in a peer-to-peer game.";
    case DPNERR_INVALIDOBJECT:
      return "The DirectPlay object pointer is invalid.";
    case DPNERR_INVALIDPARAM:
      return "One or more of the parameters passed to the method are invalid.";
    case DPNERR_INVALIDPASSWORD:
      return "An invalid password was supplied when attempting to join a session that requires a password.";
    case DPNERR_INVALIDPLAYER:
      return "The player ID is not recognized as a valid player ID for this game session.";
    case DPNERR_INVALIDPOINTER:
      return "Pointer specified as a parameter is invalid.";
    case DPNERR_INVALIDPRIORITY:
      return "The specified priority is not within the range of allowed priorities, which is inclusively from 0 through 65535.";
    case DPNERR_INVALIDSTRING:
      return "String specified as a parameter is invalid.";
    case DPNERR_INVALIDURL:
      return "Specified string is not a valid DirectPlay URL.";
    case DPNERR_INVALIDVERSION:
      return "There was an attempt to connect to an invalid version of DirectPlay.";
    case DPNERR_NOCAPS:
      return "The communication link that DirectPlay is attempting to use is not capable of this function.";
    case DPNERR_NOCONNECTION:
      return "No communication link was established.";
    case DPNERR_NOHOSTPLAYER:
      return "There is currently no player acting as the host of the session.";
    case DPNERR_NOINTERFACE:
      return "The interface is not supported.";
    case DPNERR_NORESPONSE:
      return "There was no response from the specified target.";
    case DPNERR_NOTALLOWED:
      return "Object is read-only; this function is not allowed on this object.";
    case DPNERR_NOTHOST:
      return "An attempt by the client to connect to a nonhost computer. Additionally, this error value may be returned by a nonhost that tries to set the application description.";
    case DPNERR_OUTOFMEMORY:
      return "There is insufficient memory to perform the requested operation.";
    //case DPNERR_PENDING:
      //return "Not an error, this return indicates that an asynchronous operation has reached the point where it is successfully queued. SUCCEEDED(DPNERR_PENDING) will return TRUE. This error value has been superseded by DPNERR_SUCCESS, which should be used by all new applications. DPNERR_PENDING is only included for backward compatibility.";
    case DPNERR_PLAYERLOST:
      return "A player has lost the connection to the session.";
    case DPNERR_PLAYERNOTREACHABLE:
      return "A player has tried to join a peer-peer session where at least one other existing player in the session cannot connect to the joining player.";
    case DPNERR_SESSIONFULL:
      return "The request to connect to the host or server failed because the maximum number of players allotted for the session has been reached.";
    case DPNERR_TIMEDOUT:
      return "The operation could not complete because it has timed out.";
    case DPNERR_UNINITIALIZED:
      return "The requested object has not been initialized.";
    case DPNERR_UNSUPPORTED:
      return "The function or feature is not available in this implementation or on this service provider.";
    case DPNERR_USERCANCEL:
      return "The user canceled the operation.";
    default:
      return "Unknown error.";
  }
}

NetClient::NetClient()
{ Call = NULL;
  pDPC = NULL;
}

NetClient::~NetClient()
{ Disconnect();
  if(pDPC) pDPC->Release();
}

int NetClient::Init(MSGCALL call)
{ Call = call;
  if(pDPC) return 0;

  HRESULT hRet;
  IDirectPlay8Client *dpc;

  CHKRET(DirectPlay8Create(&IID_IDirectPlay8Client, (void**)&dpc, NULL), Done);
  IFS(dpc->Initialize(this, Handler, 0)) pDPC=dpc;
  else dpc->Release();

  Done:
  return FAILED(hRet);
}

int NetClient::Send(NetMsg *msg, DWORD flags, DWORD msTimeout, DPNHANDLE *hand)
{ assert(pDPC);
  DPN_BUFFER_DESC desc;
  DPNHANDLE       dummy;
  flags |= (flags&NET_SENDGUAR ? 0 : DPNSEND_NOCOMPLETE);

  desc.dwBufferSize = msg->Len;
  desc.pBufferData  = (byte*)msg;
  return FAILED(pDPC->Send(&desc, 1, msTimeout, NULL, hand ? hand : &dummy, flags));
}

int NetClient::Connect(const char *ip)
{ assert(pDPC);
  DPN_APPLICATION_DESC   ad;
  IDirectPlay8Address   *addr, *dev;
  HRESULT                hRet;
  DPNHANDLE              dummy;
  WCHAR                  adbuf[512];

  memset(&ad, 0, sizeof(ad));
  ad.dwSize = sizeof(ad);
  ad.guidApplication = g_AppGuid;

  DirectPlay8AddressCreate(&IID_IDirectPlay8Address, (void**)&addr, NULL);
  wcscpy(adbuf, L"x-directplay:/provider=%7BEBFE7BA0-628D-11D2-AE0F-006097B01411%7D;"
                L"port=" NETPORTSTR L";hostname=");
  cw_cat(adbuf, ip);
  addr->BuildFromURLW(adbuf);

  DirectPlay8AddressCreate(&IID_IDirectPlay8Address, (void**)&dev, NULL);
  dev->BuildFromURLW(L"x-directplay:/provider=%7BEBFE7BA0-628D-11D2-AE0F-006097B01411%7D");

  hRet = pDPC->Connect(&ad, addr, dev, NULL, NULL, NULL, 0, NULL, &dummy, 0);

  addr->Release();
  dev->Release();
  return FAILED(hRet);
}

void NetClient::Disconnect()
{ if(pDPC) pDPC->Close(0);
}

HRESULT WINAPI CWorld::Handler(PVOID cls, DWORD type, PVOID msg)
{ CWorld *me = (CWorld*)cls;
  if(type==DPN_MSGID_RECEIVE)
    me->Call((NetMsg*)((DPNMSG_RECEIVE *)msg)->pReceiveData);
  else if(type==DPN_MSGID_ASYNC_OP_COMPLETE)
  { if(me->bMoveMsg == ((DPNMSG_ASYNC_OP_COMPLETE*)msg)->hAsyncOp)
      me->bMoveMsg=NULL;
  }
  else if(type==DPN_MSGID_CONNECT_COMPLETE)
  { MsgNetConn nc;
    nc.bFail = FAILED(((DPNMSG_CONNECT_COMPLETE*)msg)->hResultCode);
    me->Call(&nc);
  }
  return S_OK;
}
