#ifndef MESSAGES_H_INC
#define MESSAGES_H_INC

struct NetMsg;

#ifdef TCSERVER
  #include "..\Server\Player.h"
  typedef void(*MSGCALL)(Player *, NetMsg *);
#else
  typedef void(*MSGCALL)(NetMsg *);
#endif
#include "..\Engine\GameObj.h"
#include "..\Engine\Ammo.h"
#include "..\Engine\Weapons.h"
#include "..\Engine\World.h"

#include <pshpack1.h>

#define MSG(c,t,l)                            \
struct c : public NetMsg                      \
{ c() { Len=(l)+sizeof(NetMsg); Type=(t); }

#define END };

enum MsgType
{ MT_UNKNOWN,
  // client-to-server
  MT_CSLOGIN, MT_CSMOVE, MT_CSGETSPAWN,
  // server-to-client
  MT_SCPLAYERMOVE, MT_SCLOGINSUCC,   MT_SCLOGINFAIL, MT_SCCONNLOST,
  MT_SCPLAYERJOIN, MT_SCPLAYERLEAVE, MT_SCNETCONN,   MT_SCSPAWN,
  // either way
  MT_EWWEAPONFIRE, MT_EWWEAPONSTART, MT_EWWEAPONEND, MT_EWWEAPONSEL,
  MT_EWOBJECTHIT,  MT_EWFWEAPONHIT,  MT_EWEXPLODE,   MT_EWREMCIRCLES,
  MT_EWDEATH,      MT_EWNEWOBJECT,   MT_EWREMOBJECT
};

struct NetMsg
{ NetMsg() { Len=0; Type=MT_UNKNOWN; }
  WORD Len;
  byte Type;
};

/* CLIENT-TO-SERVER */
MSG(MsgLogin, MT_CSLOGIN, sizeof(int)+2)
  byte VerHigh, NameLen;
  int  VerLow;
  char Name[1];
END

MSG(MsgMove, MT_CSMOVE, 4*sizeof(FLOAT)+1)
  FLOAT X, Y, XV, YV;
  byte  Angle, Accel;
END

MSG(MsgGetSpawn, MT_CSGETSPAWN, 0)
END

/* SERVER-TO-CLIENT */
struct LoginPlayer
{ unsigned ID;
  WepType  Gun, Special;
  byte     Color, NameLen;
  char     Name[1];
};

MSG(MsgLoginSucc, MT_SCLOGINSUCC, sizeof(int)*5+3)
  unsigned ID, Flags;
  int      X,  Y, Color;
  byte     NumPlayers, MapLen, BackLen;
  LoginPlayer Players[1];
  // char Map[], Back[];
END

enum LoginFail
{ LF_UNKNOWN, LF_BADVERSION, LF_SERVERFULL, LF_BANNED
};
MSG(MsgLoginFail, MT_SCLOGINFAIL, 1)
  byte Reason;
END

enum ConnLost
{ CL_NETFAIL, CL_SERVERQUIT, CL_KICKED
};
MSG(MsgConnLost, MT_SCCONNLOST, 1)
  byte Reason;
END

MSG(MsgPlayerJoin, MT_SCPLAYERJOIN, sizeof(LoginPlayer)-1)
  LoginPlayer Player;
END

MSG(MsgPlayerLeave, MT_SCPLAYERLEAVE, sizeof(unsigned))
  unsigned ID;
END

MSG(MsgNetConn, MT_SCNETCONN, 1)
  bool bFail;
END

MSG(MsgPlayerMove, MT_SCPLAYERMOVE, 4*sizeof(FLOAT)+sizeof(int)+1)
  FLOAT X, Y, XV, YV;
  unsigned ID;
  byte  Angle, Accel;
END

MSG(MsgSpawn, MT_SCSPAWN, sizeof(int)*2)
  int X, Y;
END

/* EITHER WAY */
MSG(MsgWeaponFire, MT_EWWEAPONFIRE, sizeof(FLOAT)*4+sizeof(unsigned)*2+sizeof(AmmoType)+sizeof(WORD)*2+2)
  AmmoType AType;
  unsigned ID, ObjID;
  FLOAT    X, Y, XV, YV;
  WORD     Weight, Extra;
  byte     Radius, Color;
  byte     AmmoData[1];
END

MSG(MsgWeaponStart, MT_EWWEAPONSTART, sizeof(unsigned)+sizeof(WepType))
  WepType  WType;
  unsigned ID;
END

MSG(MsgWeaponEnd, MT_EWWEAPONEND, sizeof(unsigned)+sizeof(WepType))
  WepType  WType;
  unsigned ID;
END

MSG(MsgWeaponSel, MT_EWWEAPONSEL, sizeof(WepType)*2+sizeof(unsigned))
  WepType   Gun, Special;
  unsigned  ID;
END

MSG(MsgObjectHit, MT_EWOBJECTHIT, sizeof(unsigned)*3+sizeof(FLOAT)*4)
  unsigned ID, HitID, ObjID;
  FLOAT X, Y, XV, YV;
END

MSG(MsgFWeaponHit, MT_EWFWEAPONHIT, sizeof(unsigned)*2+sizeof(AmmoType)+sizeof(FLOAT)*4+sizeof(WORD)*2+1)
  AmmoType AType;
  unsigned ID, HitID;
  FLOAT X, Y, XV, YV;
  WORD  Weight, Extra;
  byte  Color;
END

MSG(MsgExplode, MT_EWEXPLODE, sizeof(int)*4+sizeof(ExpType))
  ExpType  EType;
  unsigned ID;
  int      X, Y, Seed;
END

MSG(MsgNewObject, MT_EWNEWOBJECT, sizeof(int)*4+sizeof(ObjType))
  ObjType  OType;
  unsigned ObjID, ID;
  int      X,  Y;
  byte     ObjData[1];
END

MSG(MsgRemObject, MT_EWREMOBJECT, sizeof(unsigned))
  unsigned ObjID;
END

struct CircleRem
{ int X, Y;
  byte Rad;
};
MSG(MsgRemCircles, MT_EWREMCIRCLES, sizeof(int))
  int       NumCircs;
  CircleRem Circs[1];
END

MSG(MsgDeath, MT_EWDEATH, sizeof(int))
  unsigned ID;
END

#include <poppack.h>

#undef MSG

#endif
