#include "..\Server\TCServer.h"

static const char * DPErrToString(HRESULT hRet);

/* GLOBAL */
NetServer g_Server;
NetServer::NetServer()
{ pDPS     = NULL;
  Call     = NULL;
  Spawns   = NULL;
  Players  = NULL;
  Objects  = NULL;
  NewObjID = 1;
  nPlayers = nSpawns = nCircs = maxPlayers = maxSpawns = maxCircs = 0;
  dummyPlayer = 0;
  InitializeCriticalSection(&CS);
  Init();
}

NetServer::~NetServer()
{ Close();
  Unload();
  if(pDPS) pDPS->Release();
  delete[] Circs;
  delete[] Spawns;
  delete[] Players;
  delete[] Objects;
  DeleteCriticalSection(&CS);
}

int NetServer::Init()
{ if(pDPS) return 0;
  return FAILED(DirectPlay8Create(&IID_IDirectPlay8Server, (void**)&pDPS, NULL));
}

int NetServer::Host()
{ Close();
  DPN_APPLICATION_DESC ad;
  IDirectPlay8Address *dev;
  HRESULT              hRet;

  memset(&ad, 0, sizeof(ad));
  ad.dwSize = sizeof(ad);
  ad.dwFlags = DPNSESSION_NODPNSVR | DPNSESSION_CLIENT_SERVER;
  ad.guidApplication = g_AppGuid;

  DirectPlay8AddressCreate(&IID_IDirectPlay8Address, (void**)&dev, NULL);
  dev->BuildFromURLW(L"x-directplay:/provider=%7BEBFE7BA0-628D-11D2-AE0F-006097B01411%7D;"
                     L"port=" NETPORTSTR);

  CHKRET(pDPS->Initialize(this, Handler, 0), Done);
  hRet = pDPS->Host(&ad, &dev, 1, NULL, NULL, NULL, 0);
  if(FAILED(hRet)) puts(DPErrToString(hRet));
  dev->Release();
  Done:
  return FAILED(hRet);
}

void NetServer::Close()
{ if(pDPS) pDPS->Close(0);
  dummyPlayer = 0;
}

int NetServer::Load(const char *level)
{ int   len;
  char *lev = (char*)g_VFS.LoadData(level, &len), *p, *e, *c, *orig=lev;
  if(!lev) return 1;

  Fore[0]=Back[0]=0;

  lev[len-1]=0;

  while(lev && *lev)
  { p = strchr(lev, '\n');
    if(p) *p++=0;
    g_Trim(lev);
    e = strchr(lev, '=');
    if(e)
    { *e++=0;
      if(!_stricmp(lev, "spawn"))
      { if(c=strchr(e, ','))
        { *c++=0;
          AddSpawn(atoi(e), atoi(c));
        }
      }
      else if(!stricmp(lev, "name")) SetName(e);
      else if(!stricmp(lev, "fg")) strcpy(Fore, e);
      else if(!stricmp(lev, "bg")) strcpy(Back, e);
    }
    lev=p;
  }

  delete[] orig;
  if(!Fore[0]) return 2;
  return 0;
}

void NetServer::Unload()
{ nSpawns=nCircs=0;
}

void NetServer::AddSpawn(int x, int y)
{ if(nSpawns == maxSpawns)
  { Spawn *tmp = NEW Spawn[maxSpawns+=16];
    if(Spawns)
    { memcpy(tmp, Spawns, nSpawns*sizeof(Spawn));
      delete[] Spawns;
    }
    Spawns = tmp;
  }
  Spawns[nSpawns].X = x, Spawns[nSpawns].Y = y;
  nSpawns++;
}

Player * NetServer::AddPlayer()
{ static byte Colors[] =
  { RC_LTGREEN, RC_LTRED, RC_LTBLUE, RC_LTCYAN, RC_YELLOW, RC_TAN, RC_LTMAGENTA
  };

  Player *p;
  byte    c;

  if(nPlayers == maxPlayers)
  { Player **tmp = NEW Player*[maxPlayers+=32];
    if(Players)
    { memcpy(tmp, Players, nPlayers*sizeof(Player*));
      delete[] Players;
    }
    Players = tmp;
  }
  for(int i=0; i<sizeof(Colors); i++)
  { for(int j=0; j<nPlayers; j++)
      if(Players[j]->Color==Colors[i]) break;
    if(j==nPlayers)
    { c=Colors[i];
      break;
    }
  }
  if(i==sizeof(Colors)) c = rand()%sizeof(Colors);
  p = Players[nPlayers++] = NEW Player;
  p->Color = c;
  return p;
}

void NetServer::RemPlayer(Player *p)
{ Lock();
  int i=FindPlayer(p);
  assert(i != -1);
  Players[i]=Players[--nPlayers];
  delete p;
  Unlock();
}

int NetServer::FindPlayer(Player *p)
{ for(int i=0; i<nPlayers; i++) if(Players[i]==p) return i;
  return -1;
}

void NetServer::SetName(const char *s)
{ int len=strlen(s)+1;

  delete[] Name;
  Name = NEW char[len];
  memcpy(Name, s, len);
}

int NetServer::Send(NetMsg *msg, Player *p, DWORD flags, DWORD msTimeout)
{ assert(pDPS);
  DPN_BUFFER_DESC desc;
  DPNHANDLE       dummy;
  flags |= (flags&NET_SENDGUAR ? 0 : DPNSEND_NOCOMPLETE);

  desc.dwBufferSize = msg->Len;
  desc.pBufferData  = (byte*)msg;
  return FAILED(pDPS->SendTo(p ? p->ID : DPNID_ALL_PLAYERS_GROUP,
                &desc, 1, msTimeout, NULL, &dummy, flags));
}

int NetServer::SendExc(NetMsg *msg, Player *p, DWORD flags, DWORD msTimeout)
{ int err=0;
  Lock();
  for(int i=0; i<nPlayers; i++)
    if(Players[i] != p) err+=Send(msg, Players[i], flags, msTimeout);
  Unlock();
  return err;
}

void NetServer::AddExp(MsgRemCircles *m)
{ Lock();
  int i, j, num = m->NumCircs, len = nCircs+num;
  if(len > maxCircs)
  { if(!maxCircs) maxCircs=128;
    while(maxCircs<len) maxCircs<<=1;
    CircleRem *tmp = NEW CircleRem[maxCircs];
    if(nCircs)
    { memcpy(tmp, Circs, nCircs*sizeof(CircleRem));
      delete[] Circs;
    }
    Circs = tmp;
  }
  for(i=0; i<num; i++)
  { for(j=0; j<nCircs; j++)
    { if(Circs[j].X == m->Circs[i].X && Circs[j].Y == m->Circs[i].Y &&
         Circs[j].Rad == m->Circs[i].Rad)
          break;
    }
    if(j==nCircs) Circs[nCircs++]=m->Circs[i];
  }
  Unlock();
}

void NetServer::AddObject(MsgNewObject *msg)
{
}

void NetServer::RemObject(MsgRemObject *msg)
{
}

void NetServer::MsgHandler(Player *p, NetMsg *msg)
{ union
  { NetMsg        *msg;
    MsgLogin      *login;
    MsgLoginFail  *loginfail;
    MsgLoginSucc  *loginsucc;
    MsgPlayerJoin *pjoin;
    MsgPlayerMove *pmove;
    MsgMove       *move;
    MsgWeaponSel  *wsel;
    MsgWeaponFire *wfire;
    MsgRemCircles *remcircs;
    MsgSpawn      *spawn;
    MsgNewObject  *nobj;
    MsgRemObject  *robj;
  } r, s;

  LoginPlayer *lp;
  int          i, len, num;

  char         buf[1024];
  r.msg = msg;
  s.msg = (NetMsg*)buf;
  
  if(p->State == PS_PLAYING)
  { switch(msg->Type)
    { case MT_CSMOVE:
        place_new(s.msg) MsgPlayerMove;
        s.pmove->ID = p->ID;
        s.pmove->X  = r.move->X,  s.pmove->Y  = r.move->Y;
        s.pmove->XV = r.move->XV, s.pmove->YV = r.move->YV;
        s.pmove->Angle = r.move->Angle;
        s.pmove->Accel = r.move->Accel;
        SendExc(s.pmove, p);
        break;

      case MT_EWWEAPONFIRE:
        Lock();
        r.wfire->ObjID = NewID();
        Unlock();
        Send(r.msg, NULL, NET_SENDGUAR);
        break;

      case MT_EWNEWOBJECT:
        Lock();
        r.nobj->ObjID = NewID();
        AddObject(r.nobj);
        Unlock();
        Send(r.msg, NULL, NET_SENDGUAR);
        break;

      case MT_EWREMOBJECT:
        Lock();
        RemObject(r.robj);
        Unlock();
        Send(r.msg, NULL, NET_SENDGUAR);
        break;

      case MT_EWREMCIRCLES:
        AddExp(r.remcircs);
        // fall through
      case MT_EWWEAPONSTART: case MT_EWWEAPONEND:
      case MT_EWOBJECTHIT:   case MT_EWFWEAPONHIT:
      case MT_EWDEATH:
        SendExc(r.msg, p, NET_SENDGUAR);
        break;

      case MT_EWWEAPONSEL:
        p->Gun = r.wsel->Gun, p->Special = r.wsel->Special;
        SendExc(r.msg, p, NET_SENDGUAR);
        break;

      case MT_CSGETSPAWN:
        place_new(s.spawn) MsgSpawn;
        i = rand()%nSpawns;
        s.spawn->X = Spawns[i].X, s.spawn->Y = Spawns[i].Y;
        Send(s.spawn, p);
        break; 
    }
  }
  else if(msg->Type == MT_CSLOGIN)
  { if(r.login->VerHigh != VERMAJOR || r.login->VerLow != VERMINOR)
    { place_new(s.msg) MsgLoginFail;
      s.loginfail->Reason = LF_BADVERSION;
      Send(s.loginfail, p, NET_SENDGUAR);
    }
    else
    { place_new(s.msg) MsgLoginSucc;
      p->Gun = WT_DEFGUN, p->Special = WT_DEFSPEC;
      p->SetName(r.login->Name, r.login->NameLen);
      lp = s.loginsucc->Players;
      i  = rand()%nSpawns;
      s.loginsucc->ID      = p->ID, s.loginsucc->Flags = 0;
      s.loginsucc->X       = Spawns[i].X, s.loginsucc->Y = Spawns[i].Y;
      s.loginsucc->Color   = p->Color;
      s.loginsucc->MapLen  = (i=strlen(Fore));
      s.loginsucc->BackLen = (len=strlen(Back));
      s.loginsucc->Len    += i+len;
      Lock();
      s.loginsucc->NumPlayers = nPlayers-1;
      len = 0;
      for(i=0; i<nPlayers; i++)
      { if(Players[i]==p || Players[i]->State != PS_PLAYING) continue;
        lp->ID      = Players[i]->ID;
        lp->Color   = Players[i]->Color;
        lp->Gun     = (WepType)Players[i]->Gun;
        lp->Special = (WepType)Players[i]->Special;
        lp->NameLen = (len=strlen(Players[i]->Name));
        memcpy(lp->Name, Players[i]->Name, len);
        len += sizeof(LoginPlayer)-1;
        s.loginsucc->Len += len;
        lp = (LoginPlayer*)((byte*)lp+len);
      }

      memcpy(lp, Fore, s.loginsucc->MapLen);
      lp = (LoginPlayer*)((char*)lp+s.loginsucc->MapLen);
      memcpy(lp, Back, s.loginsucc->BackLen);

      Send(s.loginsucc, p, NET_SENDGUAR);

      i = 0;
      while(i<nCircs)
      { place_new(s.msg) MsgRemCircles;
        num = min(nCircs-i, (sizeof(buf)-sizeof(MsgRemCircles))/sizeof(CircleRem));
        len = num*sizeof(CircleRem);
        memcpy(s.remcircs->Circs, Circs+i, len);
        s.remcircs->NumCircs = num;
        s.remcircs->Len     += len;
        Send(s.remcircs, p, NET_SENDGUAR);
        i += num;
      }

      Unlock();
      p->State = PS_PLAYING;

      place_new(s.msg) MsgPlayerJoin;
      s.pjoin->Player.ID      = p->ID;
      s.pjoin->Player.Color   = p->Color;
      s.pjoin->Player.Gun     = WT_DEFGUN;
      s.pjoin->Player.Special = WT_DEFSPEC;
      s.pjoin->Player.NameLen = len = strlen(p->Name);
      s.pjoin->Len += len;
      memcpy(s.pjoin->Player.Name, p->Name, s.pjoin->Player.NameLen);
      SendExc(s.pjoin, p, NET_SENDGUAR);

      printf("[Player %s(%u) logged in.]", p->Name, p->ID);
    }
  }
}

HRESULT WINAPI NetServer::Handler(PVOID cls, DWORD type, PVOID msg)
{ NetServer *me = (NetServer*)cls;
  Player *p;

  union
  { DPNMSG_CREATE_PLAYER  *create;
    DPNMSG_DESTROY_PLAYER *dest;
    DPNMSG_RECEIVE        *recv;
  } m;
    
  switch(type)
  { case DPN_MSGID_RECEIVE:
      m.recv = (DPNMSG_RECEIVE*)msg;
      if(m.recv->dpnidSender != me->dummyPlayer)
        me->MsgHandler((Player*)m.recv->pvPlayerContext, 
                       (NetMsg*)m.recv->pReceiveData);
      break;

    case DPN_MSGID_CREATE_PLAYER:
      m.create = (DPNMSG_CREATE_PLAYER*)msg;
      if(me->dummyPlayer)
      { m.create->pvPlayerContext = p = me->AddPlayer();
        p->ID = m.create->dpnidPlayer;
        printf("[Player %u connected]\nTC> ", p->ID);
      }
      else me->dummyPlayer=m.create->dpnidPlayer;
      break;

    case DPN_MSGID_DESTROY_PLAYER:
      m.dest = (DPNMSG_DESTROY_PLAYER*)msg;
      if(m.dest->dpnidPlayer != me->dummyPlayer)
      { MsgPlayerLeave msg;
        p = (Player*)m.dest->pvPlayerContext;
        msg.ID = p->ID;
        printf("[Player %s destroyed]", p->Name[0] ? p->Name : "[unknown]");
        me->RemPlayer(p);
        me->Send(&msg, NULL, NET_SENDGUAR);
      }
      break;
  }
  return S_OK;
}

const char * DPErrToString(HRESULT hRet)
{ switch(hRet)
  { case DPNSUCCESS_PENDING:
      return "An asynchronous operation has reached the point where it is successfully queued.";
    case S_OK:
      return "The operation completed successfully.";
    case DPNERR_ABORTED:
      return "The operation was canceled before it could be completed.";
    case DPNERR_ADDRESSING:
      return "The address specified is invalid.";
    case DPNERR_ALREADYCONNECTED:
      return "The object is already connected to the session.";
    case DPNERR_ALREADYCLOSING:
      return "An attempt to call the Close method on a session has been made more than once.";
    case DPNERR_ALREADYDISCONNECTING:
      return "The client is already disconnecting from the session.";
    case DPNERR_ALREADYINITIALIZED:
      return "The object has already been initialized.";
    case DPNERR_BUFFERTOOSMALL:
      return "The supplied buffer is not large enough to contain the requested data.";
    case DPNERR_CANNOTCANCEL:
      return "The operation could not be canceled.";
    case DPNERR_CANTCREATEGROUP:
      return "A new group cannot be created.";
    case DPNERR_CANTCREATEPLAYER:
      return "A new player cannot be created.";
    case DPNERR_CANTLAUNCHAPPLICATION:
      return "The lobby cannot launch the specified application.";
    case DPNERR_CONNECTING:
      return "The method is in the process of connecting to the network.";
    case DPNERR_CONNECTIONLOST:
      return "The service provider connection was reset while data was being sent.";
    case DPNERR_DATATOOLARGE:
      return "The application data is too large for the service provider's Maximum Transmission Unit.";
    case DPNERR_DOESNOTEXIST:
      return "Requested element is not part of the address.";
    case DPNERR_ENUMQUERYTOOLARGE:
      return "The query data specified is too large.";
    case DPNERR_ENUMRESPONSETOOLARGE:
      return "The response to an enumeration query is too large.";
    case DPNERR_EXCEPTION:
      return "An exception occurred when processing the request.";
    case DPNERR_GENERIC:
      return "An undefined error condition occurred.";
    case DPNERR_GROUPNOTEMPTY:
      return "The specified group is not empty.";
    case DPNERR_HOSTREJECTEDCONNECTION:
      return "The DPN_MSGID_INDICATE_CONNECT system message returned something other than S_OK in response to a connect request.";
    case DPNERR_HOSTTERMINATEDSESSION:
      return "The host in a peer session (with host migration enabled) terminated the session.";
    case DPNERR_INCOMPLETEADDRESS:
      return "The address specified is not complete.";
    case DPNERR_INVALIDADDRESSFORMAT:
      return "Address format is invalid.";
    case DPNERR_INVALIDAPPLICATION:
      return "The GUID supplied for the application is invalid.";
    case DPNERR_INVALIDCOMMAND:
      return "The command specified is invalid.";
    case DPNERR_INVALIDDEVICEADDRESS:
      return "The address for the local computer or adapter is invalid.";
    case DPNERR_INVALIDFLAGS:
      return "The flags passed to this method are invalid.";
    case DPNERR_INVALIDGROUP:
      return "The group ID is not recognized as a valid group ID for this game session.";
    case DPNERR_INVALIDHANDLE:
      return "The handle specified is invalid.";
    case DPNERR_INVALIDHOSTADDRESS:
      return "The specified remote address is invalid.";
    case DPNERR_INVALIDINSTANCE:
      return "The GUID for the application instance is invalid.";
    case DPNERR_INVALIDINTERFACE:
      return "The interface parameter is invalid. This value will be returned in a connect request if the connecting player was not a client in a client/server game or a peer in a peer-to-peer game.";
    case DPNERR_INVALIDOBJECT:
      return "The DirectPlay object pointer is invalid.";
    case DPNERR_INVALIDPARAM:
      return "One or more of the parameters passed to the method are invalid.";
    case DPNERR_INVALIDPASSWORD:
      return "An invalid password was supplied when attempting to join a session that requires a password.";
    case DPNERR_INVALIDPLAYER:
      return "The player ID is not recognized as a valid player ID for this game session.";
    case DPNERR_INVALIDPOINTER:
      return "Pointer specified as a parameter is invalid.";
    case DPNERR_INVALIDPRIORITY:
      return "The specified priority is not within the range of allowed priorities, which is inclusively from 0 through 65535.";
    case DPNERR_INVALIDSTRING:
      return "String specified as a parameter is invalid.";
    case DPNERR_INVALIDURL:
      return "Specified string is not a valid DirectPlay URL.";
    case DPNERR_INVALIDVERSION:
      return "There was an attempt to connect to an invalid version of DirectPlay.";
    case DPNERR_NOCAPS:
      return "The communication link that DirectPlay is attempting to use is not capable of this function.";
    case DPNERR_NOCONNECTION:
      return "No communication link was established.";
    case DPNERR_NOHOSTPLAYER:
      return "There is currently no player acting as the host of the session.";
    case DPNERR_NOINTERFACE:
      return "The interface is not supported.";
    case DPNERR_NORESPONSE:
      return "There was no response from the specified target.";
    case DPNERR_NOTALLOWED:
      return "Object is read-only; this function is not allowed on this object.";
    case DPNERR_NOTHOST:
      return "An attempt by the client to connect to a nonhost computer. Additionally, this error value may be returned by a nonhost that tries to set the application description.";
    case DPNERR_OUTOFMEMORY:
      return "There is insufficient memory to perform the requested operation.";
    //case DPNERR_PENDING:
      //return "Not an error, this return indicates that an asynchronous operation has reached the point where it is successfully queued. SUCCEEDED(DPNERR_PENDING) will return TRUE. This error value has been superseded by DPNERR_SUCCESS, which should be used by all new applications. DPNERR_PENDING is only included for backward compatibility.";
    case DPNERR_PLAYERLOST:
      return "A player has lost the connection to the session.";
    case DPNERR_PLAYERNOTREACHABLE:
      return "A player has tried to join a peer-peer session where at least one other existing player in the session cannot connect to the joining player.";
    case DPNERR_SESSIONFULL:
      return "The request to connect to the host or server failed because the maximum number of players allotted for the session has been reached.";
    case DPNERR_TIMEDOUT:
      return "The operation could not complete because it has timed out.";
    case DPNERR_UNINITIALIZED:
      return "The requested object has not been initialized.";
    case DPNERR_UNSUPPORTED:
      return "The function or feature is not available in this implementation or on this service provider.";
    case DPNERR_USERCANCEL:
      return "The user canceled the operation.";
    default:
      return "Unknown error.";
  }
}
