#include "..\Server\TCServer.h"
#include "..\Precomp.h"
#include "VFS.h"
#include "Genfile\memfile.h"
#include "Utility\Dataload.h"

/* Global */
CVFS g_VFS;

/*** CVFS IMPLEMENTATION ***/
CVFS::CVFS()
{ Dir       = NULL;
  Files     = NULL;
  numFiles  = maxFiles = 0;
}

CVFS::CVFS(const char *pak)
{ Dir       = NULL;
  Files     = NULL;
  numFiles  = maxFiles = 0;
  Init(pak);
}

CVFS::~CVFS()
{ for(int i=0;i<numFiles;i++) delete[] Files[i];
  delete[] Files;
  delete[] Dir;
  Pak.Deinit();
}

int CVFS::Init(const char *pak)
{ char path[_MAX_PATH];
  int  len;

  assert(!Dir);

  strcpy(path, g_Win.GetPath());
  strcat(path, BASEDATA);
  strcat(path, pak);
  strcat(path, DIRSEP);
  len = strlen(path);
  Dir = NEW char[len+1];
  strcpy(Dir, path);

  AddFiles(path);
  qsort(Files, numFiles, sizeof(char*), Compare);

  strcpy(path+len-sizeof(DIRSEP), ".pak");
  return Pak.Init(path);
}

byte * CVFS::LoadData(const char *name, int *len) const
{ assert(Dir);
  char path[_MAX_PATH];
  strcpy(path, Dir);
  strcat(path, name);
  FILE *file = fopen(path, "rb");
  byte *dat;
  if(file)
  { int   read, flen, fpos=0;
    fseek(file, 0, SEEK_END);
    flen = ftell(file);
    fseek(file, 0, SEEK_SET);
    dat  = NEW byte[flen];
    if(!dat) return NULL;
    if(len) *len = flen;
    while(read=fread(dat+fpos, 1, flen-fpos, file)) fpos+=read;
    return dat;
  }
  else
  { strcpy(path, name);
    return Pak.LoadData(path, len);
  }
}

int CVFS::LoadDataB(const char *name, memfile *file) const
{ int   len;
  byte *dat = LoadData(name, &len);
  if(!dat) return 1;
  if(file->open(dat, len, file->RW|file->FREEMEM)) return 2;
  return 0;
}

#ifndef TCSERVER
CDDSurface * CVFS::LoadPcx(const char *name, CDDSurface *dest, CDDSurfLoc loc,
                           byte *pal, CDDDepth dep) const
{ memfile file;
  if(LoadDataB(name, &file)) return NULL;
  GameLib::Loader.SetCDDraw(&g_DD);
  return GameLib::Loader.LoadPcx(&file, dest, loc, dep, pal);
}
#endif

int CVFS::GetID(const char *name) const
{ int id=FindFile(name);
  if(id==-1)
  { if((id=Pak.FindNode(name))==-1) return -1;
    return id+numFiles;
  }
  return id;
}

void CVFS::AddFiles(const char *path)
{ char buf[_MAX_PATH];
  WIN32_FIND_DATA data;
  HANDLE hand;
  int    len=strlen(path), nlen;
  strcpy(buf, path);
  strcat(buf+len, "*.*");

  if((hand=FindFirstFileA(buf, &data))==INVALID_HANDLE_VALUE) return;
  for(;;)
  { if(data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
    { if(data.cFileName[0] != '.')
      { strcpy(buf+len, data.cFileName);
        AddFiles(buf);
      }
    }
    else
    { if(numFiles==maxFiles)
      { char **tmp=NEW char*[maxFiles=maxFiles?maxFiles*2:32];
        if(Files)
        { memcpy(tmp, Files, numFiles*sizeof(char*));
          delete[] Files;
        }
        Files=tmp;
      }
      Files[numFiles]=NEW char[nlen=strlen(data.cFileName)+1];
      memcpy(Files[numFiles], data.cFileName, nlen);
      _strupr(Files[numFiles++]);
    }
    if(!FindNextFile(hand, &data)) break;
  }
}

int CVFS::FindFile(const char *name) const
{ int min=0, max=numFiles, i=max>>1, c;
  char buf[_MAX_PATH];
  if(!numFiles) return -1;

  strcpy(buf, name);
  _strupr(buf);

  while((c=strcmp(buf, Files[i])) && min<max)
  { if(c<0) max = i-1; else min = i+1;
    i = (min+max)>>1;
  }
  return c ? -1 : i;
}

int CVFS::Compare(const void *v1, const void *v2)
{ return strcmp(*(char**)v1, *(char**)v2);
}
