#ifndef VFS_H_INC
#define VFS_H_INC

#include "IntrFace\CPak.h"
namespace GameLib
{ class memfile;
}

class CVFS
{ public:
    CVFS();
    CVFS(const char *pak);
    ~CVFS();

    /* initializes this virtual file system. it takes the name of the pak file
       and subdirectory in the data directory. the data path is relative to
       the game .exe path, and is set in Frame\Defines.h. If the data path is
       "data", and the game executable resides in c:\game, it will look in 
       c:\game\data\ for the data files. If the name passed to 'pak' is
       "sound", it will attempt to load "sound.pak", which can be overridden
       by files in c:\game\data\sound. returns 0 on success
    */
    int    Init(const char *pak);
    // load data from the VFS and return a pointer to it
    byte * LoadData(const char *name, int *len) const; // len can be null
    // load data into a memory binfile
    int    LoadDataB(const char *name, GameLib::memfile *) const;

    /* load a .pcx file from the VFS. if a CDDSurface is passed into the
       function, it will use it as the destination (reinitializing it first).
       otherwise, it will create a surface using 'new'. The palette will be
       used to retrieve the palette if it exists (8-bit pcx's only). The
       pcx loader can convert between bit depths. return NULL on failure */
    GameLib::CDDSurface * LoadPcx(const char *name, GameLib::CDDSurface *,
                                  GameLib::CDDSurfLoc=GameLib::CDD_DONTCARE,
                                  byte *pal=NULL,
                                  GameLib::CDDDepth=GameLib::CDD_SAMEDEPTH)
                                  const;

    // return a unique (per vfs) ID representing each file, with -1 meaning
    // the file does not exist
    int    GetID(const char *name) const;

    // returns the number of entries in the underlying .pak. does NOT take
    // into account any standalone files that may in the .pak
    int    NumEntries() const { return Pak.NumEntries()+numFiles; }

  private:
    void   AddFiles(const char *path);
    int    FindFile(const char *name) const;
    static int Compare(const void *, const void *);
    GameLib::CPak   Pak;
    char           *Dir;
    char          **Files;
    int             numFiles, maxFiles;
};

#endif
